# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/9/2 15:35
from enum import Enum
from functools import wraps
from typing import Optional

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, FileStorage, RawModel

app = OpenAPI(__name__)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


class TypeEnum(str, Enum):
    A = "A"
    B = "B"


class BookForm(BaseModel):
    file: FileStorage
    files: list[FileStorage]
    string: str
    string_list: list[str]


class BookQuery(BaseModel):
    age: list[int]
    book_type: Optional[TypeEnum] = None


class BookBody(BaseModel):
    age: int


class BookCookie(BaseModel):
    token: Optional[str] = None
    token_type: Optional[TypeEnum] = None


class BookHeader(BaseModel):
    Hello1: str = Field("what's up", max_length=12, description="sds")
    # required
    hello2: str = Field(..., max_length=12, description="sds")
    api_key: str = Field(..., description="API Key")
    api_type: Optional[TypeEnum] = None
    x_hello: str = Field(..., max_length=12, description='Header with alias to support dash', alias="x-hello")


def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@app.get("/query")
@decorator
def api_query(query: BookQuery):
    print(query)
    return {"code": 0, "message": "ok"}


@app.post("/form")
def api_form(form: BookForm):
    print(form)
    return {"code": 0, "message": "ok"}


@app.post("/body")
def api_error_json(body: BookBody):
    print(body)  # pragma: no cover


@app.get("/header")
def get_book(header: BookHeader):
    return header.model_dump(by_alias=True)


@app.post("/cookie")
def api_cookie(cookie: BookCookie):
    print(cookie)
    return {"code": 0, "message": "ok"}


class BookRaw(RawModel):
    mimetypes = ["text/csv", "application/json"]


@app.post("/raw")
def api_raw(raw: BookRaw):
    # raw equals to flask.request
    assert raw.data == b"raw"
    assert raw.mimetype == "text/plain"
    return "ok"


def test_query(client):
    r = client.get("/query?age=1")
    print(r.json)
    assert r.status_code == 200


def test_form(client):
    from io import BytesIO
    data = {
        "file": (BytesIO(b"post-data"), "filename"),
        "files": [(BytesIO(b"post-data"), "filename"), (BytesIO(b"post-data"), "filename")],
        "string": "a",
        "string_list": ["a", "b", "c"]
    }
    r = client.post("/form", data=data, content_type="multipart/form-data")
    assert r.status_code == 200


def test_error_json(client):
    r = client.post("/body", json="{age: 1}")
    assert r.status_code == 422


def test_cookie(client):
    r = client.post("/cookie")
    print(r.json)
    assert r.status_code == 200


def test_header(client):
    headers = {"Hello1": "111", "hello2": "222", "api_key": "333", "api_type": "A", "x-hello": "444"}
    resp = client.get("/header", headers=headers)
    print(resp.json)
    assert resp.status_code == 200
    assert resp.json == headers


def test_raw(client):
    resp = client.post("/raw", data="raw", headers={"Content-Type": "text/plain"})
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/6 16:38
from __future__ import annotations

import json
from http import HTTPStatus
from typing import Optional

import pytest
from flask import Response
from pydantic import BaseModel, RootModel, Field

from flask_openapi3 import ExternalDocumentation
from flask_openapi3 import Info, Tag
from flask_openapi3 import OpenAPI

info = Info(title='book API', version='1.0.0')

jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
security_schemes = {"jwt": jwt}


class NotFoundResponse(BaseModel):
    code: int = Field(-1, description="Status Code")
    message: str = Field("Resource not found!", description="Exception Information")


def get_operation_id_for_path_callback(*, name: str, path: str, method: str) -> str:
    print(name, path, method)
    return name


app = OpenAPI(
    __name__,
    info=info,
    security_schemes=security_schemes,
    responses={"404": NotFoundResponse},
    operation_id_callback=get_operation_id_for_path_callback,
)
app.config["TESTING"] = True

security = [{"jwt": []}]
book_tag = Tag(name='book', description='Book')


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description='Age')


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


class BookPath(BaseModel):
    bid: int = Field(..., description='book id')


class BookBodyWithID(BaseModel):
    bid: int = Field(..., description='book id')
    age: Optional[int] = Field(None, ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


class BaseResponse(BaseModel):
    code: int = Field(0, description="Status Code")
    message: str = Field("ok", description="Exception Information")


class BookListResponseV1(BaseResponse):
    data: list[BookBodyWithID] = Field(..., description="All the books")


class BookListResponseV2(BaseModel):
    books: list[BookBodyWithID] = Field(...)


class BookListResponseV3(RootModel):
    root: list[BookBodyWithID]


class BookResponse(BaseModel):
    code: int = Field(0, description="Status Code")
    message: str = Field("ok", description="Exception Information")
    data: BookBodyWithID


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.get(
    '/book/<int:bid>',
    tags=[book_tag],
    operation_id="get_book_id",
    external_docs=ExternalDocumentation(
        url="https://www.openapis.org/",
        description="Something great got better, get excited!"),
    responses={"200": BookResponse},
    security=security
)
def get_book(path: BookPath):
    """Get a book
    to get some book by id, like:
    http://localhost:5000/book/3
    """
    if path.bid == 4:
        return NotFoundResponse().model_dump(), 404


@app.get('/book', tags=[book_tag], responses={"200": BookListResponseV1})
def get_books(query: BookBody):
    """get books
    to get all books
    """
    assert query.age == 3
    assert query.author == 'joy'
    return {
        "code": 0,
        "message": "ok",
        "data": [
            {"bid": 1, "age": query.age, "author": "b1"},
            {"bid": 2, "age": query.age, "author": "b2"}
        ]
    }


@app.get('/book_v2', tags=[book_tag], responses={"200": BookListResponseV2})
def get_books_v2(query: BookBody):
    """get books
    to get all books (v2)
    """
    assert query.age == 3
    assert query.author == 'joy'
    return {
        "books": [
            {"bid": 1, "age": query.age, "author": "b1"},
            {"bid": 2, "age": query.age, "author": "b2"}
        ]
    }


@app.get('/book_v3', tags=[book_tag], responses={"200": BookListResponseV3})
def get_books_v3(query: BookBody):
    """get books
    to get all books (v3)
    """
    assert query.age == 3
    assert query.author == 'joy'

    books = [
        {"bid": 1, "age": query.age, "author": "b1"},
        {"bid": 2, "age": query.age, "author": "b2"}
    ]
    # A `list` have to be converted to json-format `str` returned as a `Response` object,
    # because flask doesn't support returning a `list` as a response
    return Response(json.dumps(books), status=200, headers={'Content-Type': 'application/json'})


@app.post('/book', tags=[book_tag], responses={"200": BaseResponse})
def create_book(body: BookBody):
    assert body.age == 3
    return {"code": 0, "message": "ok"}, HTTPStatus.OK


@app.put('/book/<int:bid>', tags=[book_tag])
def update_book(path: BookPath, body: BookBody):
    assert path.bid == 1
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@app.patch('/book/<int:bid>', tags=[book_tag], doc_ui=False)
def update_book1(path: BookPath, body: BookBody):
    assert path.bid == 1
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@app.delete('/book/<int:bid>', tags=[book_tag], responses={"200": BaseResponse})
def delete_book(path: BookPath):
    assert path.bid == 1
    return {"code": 0, "message": "ok"}


@app.delete('/book_no_response/<int:bid>', tags=[book_tag], responses={"204": None})
def delete_book_no_response(path: BookPath):
    assert path.bid == 1
    return b'', 204


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    print(resp.json)
    assert resp.status_code == 200
    assert resp.json == app.api_doc


def test_get(client):
    resp = client.get("/book?age=3&author=joy")
    assert resp.status_code == 200
    resp_v2 = client.get("/book_v2?age=3&author=joy")
    assert resp_v2.status_code == 200
    resp_v3 = client.get("/book_v3?age=3&author=joy")
    assert resp_v3.status_code == 200


def test_get_by_id_4(client):
    resp = client.get("/book/4?age=3&author=joy")
    assert resp.status_code == 404


def test_post(client):
    resp = client.post("/book", json={"age": 3})
    assert resp.status_code == 200


def test_put(client):
    resp = client.put("/book/1", json={"age": 3})
    assert resp.status_code == 200


def test_patch(client):
    resp = client.patch("/book/1", json={"age": 3})
    assert resp.status_code == 200


def test_delete(client):
    resp = client.delete("/book/1")
    assert resp.status_code == 200


def test_delete_no_response(client):
    resp = client.delete("/book_no_response/1")
    assert resp.status_code == 204



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/12/1 9:39

import pytest
from pydantic import BaseModel

from flask_openapi3 import Info, OpenAPI

info = Info(title='book API', version='1.0.0')

app = OpenAPI(__name__, info=info)
app.config["TESTING"] = True


class CreateBookBody(BaseModel):
    pass

    model_config = {
        "extra": "allow",
    }


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.post('/book')
def create_book(body: CreateBookBody):
    print(body.model_dump())
    return {"code": 0, "message": "ok"}


def test_post(client):
    resp = client.post("/book", json={'aaa': 111, 'bbb': 222})
    print(resp.json)
    assert resp.status_code == 200



from __future__ import annotations

from typing import Generic, TypeVar, Optional, Literal

from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, Schema

T = TypeVar("T", bound=BaseModel)


def test_responses_are_replicated_in_open_api(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    class BaseResponse(BaseModel):
        """Base description"""
        test: int

        model_config = dict(
            openapi_extra={
                "description": "Custom description",
                "headers": {
                    "location": {
                        "description": "URL of the new resource",
                        "schema": {"type": "string"}
                    }
                },
                "content": {
                    "text/plain": {
                        "schema": {"type": "string"}
                    }
                },
                "links": {
                    "dummy": {
                        "description": "dummy link"
                    }
                }
            }
        )

    @test_app.get("/test", responses={"201": BaseResponse})
    def endpoint_test():
        return b"", 201  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["get"]["responses"]["201"] == {
            "description": "Custom description",
            "headers": {
                "location": {
                    "description": "URL of the new resource",
                    "schema": {"type": "string"}
                }
            },
            "content": {
                # This content is coming from responses
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/BaseResponse"}
                },
                # While this one comes from responses
                "text/plain": {
                    "schema": {"type": "string"}
                }
            },
            "links": {
                "dummy": {
                    "description": "dummy link"
                }
            }
        }


def test_none_responses_are_replicated_in_open_api(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.get(
        "/test",
        responses={
            "204": {
                "description": "Custom description",
                "headers": {
                    "x-my-special-header": {
                        "description": "Custom header",
                        "schema": {"type": "string"}
                    }
                },
                "content": {
                    "text/plain": {
                        "schema": {"type": "string"}
                    }
                },
                "links": {
                    "dummy": {
                        "description": "dummy link"
                    }
                }
            }
        }
    )
    def endpoint_test():
        return b"", 204  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["get"]["responses"]["204"] == {
            "description": "Custom description",
            "headers": {
                "x-my-special-header": {
                    "description": "Custom header",
                    "schema": {"type": "string"}
                }
            },
            "content": {
                "text/plain": {
                    "schema": {"type": "string"}
                }
            },
            "links": {
                "dummy": {
                    "description": "dummy link"
                }
            }
        }


def test_responses_are_replicated_in_open_api2(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.get(
        "/test",
        responses={
            "201": {
                "description": "Custom description",
                "headers": {
                    "location": {
                        "description": "URL of the new resource",
                        "schema": {"type": "string"}
                    }
                },
                "content": {
                    "text/plain": {
                        "schema": {"type": "string"}
                    }
                },
                "links": {
                    "dummy": {
                        "description": "dummy link"
                    }
                }
            }
        }
    )
    def endpoint_test():
        return b"", 201  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["get"]["responses"]["201"] == {
            "description": "Custom description",
            "headers": {
                "location": {
                    "description": "URL of the new resource",
                    "schema": {"type": "string"}
                }
            },
            "content": {
                "text/plain": {
                    "schema": {"type": "string"}
                }
            },
            "links": {
                "dummy": {
                    "description": "dummy link"
                }
            }
        }


def test_responses_without_content_are_replicated_in_open_api(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.get(
        "/test",
        responses={
            "201": {
                "description": "Custom description",
                "headers": {
                    "location": {
                        "description": "URL of the new resource",
                        "schema": {"type": "string"}
                    }
                },
                "links": {
                    "dummy": {
                        "description": "dummy link"
                    }
                }
            }
        }
    )
    def endpoint_test():
        return b"", 201  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["get"]["responses"]["201"] == {
            "description": "Custom description",
            "headers": {
                "location": {
                    "description": "URL of the new resource",
                    "schema": {"type": "string"}
                }
            },
            "links": {
                "dummy": {
                    "description": "dummy link"
                }
            }
        }


class BaseRequest(BaseModel):
    """Base description"""
    test_int: int
    test_str: str


class BaseRequestGeneric(BaseModel, Generic[T]):
    detail: T

    model_config = dict(
        openapi_extra={
            "examples": {
                "Example 01": {
                    "summary": "An example",
                    "value": {
                        "test_int": -1,
                        "test_str": "negative",
                    }
                },
                "Example 02": {
                    "externalValue": "https://example.org/examples/second-example.xml"
                },
                "Example 03": {
                    "$ref": "#/components/examples/third-example"
                }
            }
        }
    )


def test_body_examples_are_replicated_in_open_api(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(body: BaseRequestGeneric[BaseRequest]):
        return body.model_dump(), 200

    with test_app.test_client() as client:
        client.post("/test", json={"detail": {"test_int": 1, "test_str": "s"}})
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["post"]["requestBody"] == {
            "content": {
                "application/json": {
                    "examples": {
                        "Example 01": {"summary": "An example", "value": {"test_int": -1, "test_str": "negative"}},
                        "Example 02": {"externalValue": "https://example.org/examples/second-example.xml"},
                        "Example 03": {"$ref": "#/components/examples/third-example"}
                    },
                    "schema": {"$ref": "#/components/schemas/BaseRequestGeneric_BaseRequest_"}
                }
            },
            "required": True
        }
        assert resp.json["components"]["schemas"]["BaseRequestGeneric_BaseRequest_"] == {
            "properties": {"detail": {"$ref": "#/components/schemas/BaseRequest"}},
            "required": ["detail"],
            "title": "BaseRequestGeneric[BaseRequest]",
            "type": "object",
        }


def test_form_examples(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    model_config = dict(
        openapi_extra={
            "examples": {
                "Example 01": {
                    "summary": "An example",
                    "value": {
                        "test_int": -1,
                        "test_str": "negative",
                    }
                }
            }
        }
    )
    BaseRequestGeneric[BaseRequest].model_config = model_config

    @test_app.post("/test")
    def endpoint_test(form: BaseRequestGeneric[BaseRequest]):
        return form.model_dump(), 200  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["post"]["requestBody"] == {
            "content": {
                "multipart/form-data": {
                    "schema": {"$ref": "#/components/schemas/BaseRequestGeneric_BaseRequest_"},
                    "examples": {
                        "Example 01": {"summary": "An example", "value": {"test_int": -1, "test_str": "negative"}}
                    }
                }
            },
            "required": True
        }
        assert resp.json["components"]["schemas"]["BaseRequestGeneric_BaseRequest_"] == {
            "properties": {"detail": {"$ref": "#/components/schemas/BaseRequest"}},
            "required": ["detail"],
            "title": "BaseRequestGeneric[BaseRequest]",
            "type": "object",
        }


class BaseRequestBody(BaseModel):
    base: BaseRequest


def test_body_with_complex_object(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(body: BaseRequestBody):
        return body.model_dump(), 200  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert {"properties", "required", "title", "type"} == set(
            resp.json["components"]["schemas"]["BaseRequestBody"].keys())


class Detail(BaseModel):
    num: int


class GenericResponse(BaseModel, Generic[T]):
    detail: T


class ListGenericResponse(BaseModel, Generic[T]):
    items: list[GenericResponse[T]]


def test_responses_with_generics(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.get("/test", responses={"201": ListGenericResponse[Detail]})
    def endpoint_test():
        return b"", 201  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["get"]["responses"]["201"] == {
            "description": "Created",
            "content": {
                "application/json": {
                    "schema": {"$ref": "#/components/schemas/ListGenericResponse_Detail_"}
                },
            },
        }

        schemas = resp.json["components"]["schemas"]
        detail = schemas["ListGenericResponse_Detail_"]
        assert detail["title"] == "ListGenericResponse[Detail]"
        assert detail["properties"]["items"]["items"]["$ref"] == "#/components/schemas/GenericResponse_Detail_"
        assert schemas["GenericResponse_Detail_"]["title"] == "GenericResponse[Detail]"


class PathParam(BaseModel):
    type_name: str = Field(..., description="id for path", max_length=300,
                           json_schema_extra={"deprecated": False, "example": "42"})


def test_path_parameter_object(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(path: PathParam):
        return path.model_dump(), 200  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["post"]["parameters"][0] == {
            "deprecated": False,
            "description": "id for path",
            "example": "42",
            "in": "path",
            "name": "type_name",
            "required": True,
            "schema": {
                "deprecated": False,
                "description": "id for path",
                "maxLength": 300,
                "example": "42",
                "title": "Type Name",
                "type": "string",
            },
        }


class QueryParam(BaseModel):
    count: int = Field(..., description="count of param", le=1000.0,
                       json_schema_extra={"deprecated": True, "example": 100})


def test_query_parameter_object(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(query: QueryParam):
        return query.model_dump(), 200  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["post"]["parameters"][0] == {
            "deprecated": True,
            "description": "count of param",
            "example": 100,
            "in": "query",
            "name": "count",
            "required": True,
            "schema": {
                "deprecated": True,
                "description": "count of param",
                "maximum": 1000.0,
                "example": 100,
                "title": "Count",
                "type": "integer",
            },
        }


class HeaderParam(BaseModel):
    app_name: str = Field(None, description="app name")


class CookieParam(BaseModel):
    app_name: str = Field(None, description="app name", json_schema_extra={"example": "aaa"})


def test_header_parameter_object(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(header: HeaderParam, cookie: CookieParam):
        print(header, cookie)  # pragma: no cover

    with test_app.test_client() as client:
        resp = client.get("/openapi/openapi.json")
        assert resp.status_code == 200
        assert resp.json["paths"]["/test"]["post"]["parameters"][0] == {
            "description": "app name",
            "in": "header",
            "name": "app_name",
            "required": False,
            "schema": {"description": "app name", "title": "App Name", "type": "string", "default": None}
        }
        assert resp.json["paths"]["/test"]["post"]["parameters"][1] == {
            "description": "app name",
            "in": "cookie",
            "example": "aaa",
            "name": "app_name",
            "required": False,
            "schema": {"description": "app name", "example": "aaa", "title": "App Name", "type": "string",
                       "default": None}
        }


class Model(BaseModel):
    one: Optional[int] = Field(default=None)
    two: Optional[int] = Field(default=2)


def test_default_none(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(body: Model):
        print([])  # pragma: no cover

    works = Model.model_json_schema()["properties"]
    assert works["one"]["default"] is None
    assert works["two"]["default"] == 2
    breaks = test_app.api_doc["components"]["schemas"]["Model"]["properties"]
    assert breaks["two"]["default"] == 2
    assert breaks["one"]["default"] is None


def test_parameters_none(request):
    """Parameters key shouldn't exist."""
    test_app = OpenAPI(request.node.name)

    @test_app.post("/test")
    def endpoint_test():
        print([])  # pragma: no cover

    data = test_app.api_doc["paths"]["/test"]["post"]

    assert "parameters" not in data


def test_deprecated_none(request):
    """Parameters key shouldn't exist."""
    test_app = OpenAPI(request.node.name)

    @test_app.post("/test")
    def endpoint_test():
        print([])  # pragma: no cover

    data = test_app.api_doc["paths"]["/test"]["post"]

    assert "deprecated" not in data


class TupleModel(BaseModel):
    my_tuple: tuple[Literal["a", "b"], Literal["c", "d"]]


def test_prefix_items(request):
    test_app = OpenAPI(request.node.name)
    test_app.config["TESTING"] = True

    @test_app.post("/test")
    def endpoint_test(body: TupleModel):
        print([])  # pragma: no cover

    schema = test_app.api_doc["paths"]["/test"]["post"]["requestBody"]["content"]["application/json"]["schema"]
    assert schema == {'$ref': '#/components/schemas/TupleModel'}
    components = test_app.api_doc["components"]["schemas"]
    assert components["TupleModel"] == {'properties': {'my_tuple': {'maxItems': 2,
                             'minItems': 2,
                             'prefixItems': [{'enum': ['a', 'b'],
                                              'type': 'string'},
                                             {'enum': ['c', 'd'],
                                              'type': 'string'}],
                             'title': 'My Tuple',
                             'type': 'array'}},
 'required': ['my_tuple'],
 'title': 'TupleModel',
 'type': 'object'}


def test_schema_bigint(request):
    max_nr = 9223372036854775807
    obj = Schema(maximum=max_nr)
    assert obj.model_dump()["maximum"] == max_nr



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/10/14 16:09
import typing
from typing import Optional, Any, Callable

from .models import ExternalDocumentation
from .models import Server
from .models import Tag
from .types import ResponseDict
from .utils import HTTPMethod
from .utils import convert_responses_key_to_string
from .utils import get_operation
from .utils import get_operation_id_for_path
from .utils import get_responses
from .utils import parse_and_store_tags
from .utils import parse_method
from .utils import parse_parameters
from .utils import parse_rule

if typing.TYPE_CHECKING:  # pragma: no cover
    from .openapi import OpenAPI


class APIView:
    def __init__(
            self,
            url_prefix: Optional[str] = None,
            view_tags: Optional[list[Tag]] = None,
            view_security: Optional[list[dict[str, list[str]]]] = None,
            view_responses: Optional[ResponseDict] = None,
            doc_ui: bool = True,
            operation_id_callback: Callable = get_operation_id_for_path,
    ):
        """
        Create a class-based view

        Args:
            url_prefix: A path to prepend to all the APIView's urls
            view_tags: APIView tags for every API.
            view_security: APIView security for every API.
            view_responses: API responses should be either a subclass of BaseModel, a dictionary, or None.
            doc_ui: Enable OpenAPI document UI (Swagger UI and Redoc). Defaults to True.
            operation_id_callback: Callback function for custom operation_id generation.
                                   Receives name (str), path (str) and method (str) parameters.
                                   Defaults to `get_operation_id_for_path` from utils
        """
        self.url_prefix = url_prefix
        self.view_tags = view_tags or []
        self.view_security = view_security or []

        # Convert key to string
        self.view_responses = convert_responses_key_to_string(view_responses or {})

        self.doc_ui = doc_ui
        self.operation_id_callback: Callable = operation_id_callback

        self.views: dict = dict()
        self.paths: dict = dict()
        self.components_schemas: dict = dict()
        self.tags: list[Tag] = []
        self.tag_names: list[str] = []

    def route(self, rule: str):
        """Decorator for view class"""

        def wrapper(cls):
            if self.views.get(rule):  # pragma: no cover
                raise ValueError(f"malformed url rule: {rule!r}")
            methods = []

            # Parse rule: merge url_prefix and format rule from /pet/<petId> to /pet/{petId}
            uri = parse_rule(rule, url_prefix=self.url_prefix)

            for method in HTTPMethod:
                cls_method = getattr(cls, method.lower(), None)
                if not cls_method:
                    continue
                methods.append(method)
                if self.doc_ui is False:
                    continue
                if not getattr(cls_method, "operation", None):
                    continue
                # Parse method
                parse_method(uri, method, self.paths, cls_method.operation)
                # Update operation_id
                if not cls_method.operation.operationId:
                    cls_method.operation.operationId = self.operation_id_callback(
                        name=cls_method.__qualname__,
                        path=rule,
                        method=method
                    )

            # Convert route parameters from {param} to <param>
            _rule = uri.replace("{", "<").replace("}", ">")
            self.views[_rule] = (cls, methods)

            return cls

        return wrapper

    def doc(
            self,
            *,
            tags: Optional[list[Tag]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            external_docs: Optional[ExternalDocumentation] = None,
            operation_id: Optional[str] = None,
            responses: Optional[ResponseDict] = None,
            deprecated: Optional[bool] = None,
            security: Optional[list[dict[str, list[Any]]]] = None,
            servers: Optional[list[Server]] = None,
            openapi_extensions: Optional[dict[str, Any]] = None,
            doc_ui: bool = True
    ) -> Callable:
        """
        Decorator for view method.
        More information goto https://spec.openapis.org/oas/v3.1.0#operation-object

        Args:
            tags: Adds metadata to a single tag.
            summary: A short summary of what the operation does.
            description: A verbose explanation of the operation behavior.
            external_docs: Additional external documentation for this operation.
            operation_id: Unique string used to identify the operation.
            responses: API responses should be either a subclass of BaseModel, a dictionary, or None.
            deprecated: Declares this operation to be deprecated.
            security: A declaration of which security mechanisms can be used for this operation.
            servers: An alternative server array to service this operation.
            openapi_extensions: Allows extensions to the OpenAPI Schema.
            doc_ui: Declares this operation to be shown. Default to True.
        """

        new_responses = convert_responses_key_to_string(responses or {})
        security = security or []
        tags = tags + self.view_tags if tags else self.view_tags

        def decorator(func):
            if self.doc_ui is False or doc_ui is False:
                return func

            # Global response combines API responses
            combine_responses = {**self.view_responses, **new_responses}

            # Create operation
            operation = get_operation(
                func,
                summary=summary,
                description=description,
                openapi_extensions=openapi_extensions
            )

            # Set external docs
            if external_docs:
                operation.externalDocs = external_docs

            # Unique string used to identify the operation.
            if operation_id:
                operation.operationId = operation_id

            # Only set `deprecated` if True, otherwise leave it as None
            if deprecated is not None:
                operation.deprecated = deprecated

            # Add security
            _security = (security or []) + self.view_security or None
            if _security:
                operation.security = _security

            # Add servers
            if servers:
                operation.servers = servers

            # Store tags
            parse_and_store_tags(tags, self.tags, self.tag_names, operation)

            # Parse parameters
            parse_parameters(
                func,
                components_schemas=self.components_schemas,
                operation=operation
            )

            # Parse response
            get_responses(combine_responses, self.components_schemas, operation)
            func.operation = operation

            return func

        return decorator

    def register(
            self,
            app: "OpenAPI",
            url_prefix: Optional[str] = None,
            view_kwargs: Optional[dict[Any, Any]] = None
    ) -> None:
        """
        Register the API views with the given OpenAPI app.

        Args:
            app: An instance of the OpenAPI app.
            url_prefix: A path to prepend to all the APIView's urls
            view_kwargs: Additional keyword arguments to pass to the API views.
        """
        for rule, (cls, methods) in self.views.items():
            for method in methods:
                func = getattr(cls, method.lower())
                header, cookie, path, query, form, body, raw = parse_parameters(func, doc_ui=False)
                view_func = app.create_view_func(
                    func,
                    header,
                    cookie,
                    path,
                    query,
                    form,
                    body,
                    raw,
                    view_class=cls,
                    view_kwargs=view_kwargs
                )

                if url_prefix and self.url_prefix and url_prefix != self.url_prefix:
                    rule = url_prefix + rule.removeprefix(self.url_prefix)
                elif url_prefix and not self.url_prefix:
                    rule = url_prefix.rstrip("/") + "/" + rule.lstrip("/")

                options = {
                    "endpoint": cls.__name__ + "." + method.lower(),
                    "methods": [method.upper()]
                }
                app.add_url_rule(rule, view_func=view_func, **options)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:41
from typing import Optional

from pydantic import BaseModel


class Discriminator(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#discriminator-object
    """

    propertyName: str
    mapping: Optional[dict[str, str]] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/10/18 9:00
from typing import Optional

from pydantic import BaseModel, Field

from flask_openapi3 import APIView
from flask_openapi3 import OpenAPI, Tag, Info

jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
security_schemes = {"jwt": jwt}
info = Info(title='book API', version='1.0.0')
app = OpenAPI(__name__, info=info, security_schemes=security_schemes)

security = [{"jwt": []}]

api_view = APIView(url_prefix="/api/v1", view_tags=[Tag(name="book")], view_security=security)


class BookPath(BaseModel):
    id: int = Field(..., description="book ID")


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description='Age')


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


@api_view.route("/book")
class BookListAPIView:
    a = 1

    @api_view.doc(summary="get book list")
    def get(self, query: BookQuery):
        print(self.a)
        return query.model_dump_json()

    @api_view.doc(summary="create book")
    def post(self, body: BookBody):
        """description for a created book"""
        return body.model_dump_json()


@api_view.route("/book/<id>")
class BookAPIView:
    @api_view.doc(summary="get book")
    def get(self, path: BookPath):
        print(path)
        return "get"

    @api_view.doc(summary="update book")
    def put(self, path: BookPath):
        print(path)
        return "put"

    @api_view.doc(summary="delete book", deprecated=True)
    def delete(self, path: BookPath):
        print(path)
        return "delete"


app.register_api_view(api_view)

if __name__ == "__main__":
    print(app.url_map)
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/17 15:25

from typing import Optional

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import APIBlueprint, OpenAPI
from flask_openapi3 import Tag, Info

info = Info(title='book API', version='1.0.0')

jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
security_schemes = {"jwt": jwt}

app = OpenAPI(__name__, info=info, security_schemes=security_schemes)
app.config["TESTING"] = True

tag = Tag(name='book', description="Book")
security = [{"jwt": []}]


class Unauthorized(BaseModel):
    code: int = Field(-1, description="Status Code")
    message: str = Field("Unauthorized!", description="Exception Information")


api = APIBlueprint(
    '/book',
    __name__,
    url_prefix='/api',
    abp_tags=[tag],
    abp_security=security,
    abp_responses={"401": Unauthorized}
)

try:
    api.register_api(api)
except ValueError as e:
    assert str(e) == "Cannot register a api blueprint on itself"


@pytest.fixture
def client():
    client = app.test_client()

    return client


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


class BookPath(BaseModel):
    bid: int = Field(..., description='book id')


@api.post('/book', doc_ui=False)
def create_book(body: BookBody):
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@api.put('/book/<int:bid>', operation_id='update')
def update_book(path: BookPath, body: BookBody):
    assert path.bid == 1
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@api.patch('/book/<int:bid>')
def update_book1(path: BookPath, body: BookBody):
    assert path.bid == 1
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@api.patch('/v2/book/<int:bid>')
def update_book1_v2(path: BookPath, body: BookBody):
    assert path.bid == 1
    assert body.age == 3
    return {"code": 0, "message": "ok"}


@api.delete('/book/<int:bid>')
def delete_book(path: BookPath):
    assert path.bid == 1
    return {"code": 0, "message": "ok"}


# register api
app.register_api(api)


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc
    assert resp.json["paths"]["/api/book/{bid}"]["put"]["operationId"] == "update"
    assert resp.json["paths"]["/api/book/{bid}"]["delete"]["operationId"] == "_book_book__int_bid__delete"


def test_post(client):
    resp = client.post("/api/book", json={"age": 3})
    assert resp.status_code == 200


def test_put(client):
    resp = client.put("/api/book/1", json={"age": 3})
    assert resp.status_code == 200


def test_patch(client):
    resp = client.patch("/api/book/1", json={"age": 3})
    assert resp.status_code == 200

    resp = client.patch("/api/v2/book/1", json={"age": 3})
    assert resp.status_code == 200


def test_delete(client):
    resp = client.delete("/api/book/1")
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/4/28 10:58
"""
OpenAPI v3.1.0 schema types, created according to the specification:
https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md

The type orders are according to the contents of the specification:
https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#table-of-contents
"""

from typing import Optional, Union

from flask import Request
from pydantic import BaseModel

from .callback import Callback
from .components import Components
from .contact import Contact
from .discriminator import Discriminator
from .encoding import Encoding
from .example import Example
from .external_documentation import ExternalDocumentation
from .file import FileStorage
from .header import Header
from .info import Info
from .license import License
from .link import Link
from .media_type import MediaType
from .oauth_flow import OAuthFlow
from .oauth_flows import OAuthFlows
from .operation import Operation
from .parameter import Parameter
from .parameter_in_type import ParameterInType
from .path_item import PathItem
from .paths import Paths
from .reference import Reference
from .request_body import RequestBody
from .response import Response
from .responses import Responses
from .schema import Schema
from .security_requirement import SecurityRequirement
from .security_scheme import SecurityScheme
from .server import Server
from .server_variable import ServerVariable
from .style_values import StyleValues
from .tag import Tag
from .validation_error import UnprocessableEntity
from .validation_error import ValidationErrorModel
from .xml import XML

OPENAPI3_REF_PREFIX = "#/components/schemas"
OPENAPI3_REF_TEMPLATE = OPENAPI3_REF_PREFIX + "/{model}"


class APISpec(BaseModel):
    """https://spec.openapis.org/oas/v3.1.0#openapi-object"""
    openapi: str
    info: Info
    servers: Optional[list[Server]] = None
    paths: Paths
    components: Optional[Components] = None
    security: Optional[list[SecurityRequirement]] = None
    tags: Optional[list[Tag]] = None
    externalDocs: Optional[ExternalDocumentation] = None
    webhooks: Optional[dict[str, Union[PathItem, Reference]]] = None

    model_config = {
        "extra": "allow"
    }


class OAuthConfig(BaseModel):
    """
    https://github.com/swagger-api/swagger-ui/blob/master/docs/usage/oauth2.md#oauth-20-configuration
    """
    clientId: Optional[str] = None
    clientSecret: Optional[str] = None
    realm: Optional[str] = None
    appName: Optional[str] = None
    scopeSeparator: Optional[str] = None
    scopes: Optional[str] = None
    additionalQueryStringParams: Optional[dict[str, str]] = None
    useBasicAuthenticationWithAccessCodeGrant: Optional[bool] = False
    usePkceWithAuthorizationCodeGrant: Optional[bool] = False


class RawModel(Request):
    mimetypes: list[str] = ["application/json"]


Encoding.model_rebuild()
Operation.model_rebuild()
PathItem.model_rebuild()



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/6/26 17:05

from flask_openapi3 import OpenAPI
from flask import make_response

app = OpenAPI(__name__)


@app.get('/image')
def get_image():
    with open("../docs/images/openapi.png", "rb") as f:
        content = f.read()

    r = make_response(content)
    r.mimetype = 'image/png'
    return r


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/9/12 17:19
from flask_openapi3 import OpenAPI
from flask_openapi3 import RawModel

app = OpenAPI(__name__)


class BookRaw(RawModel):
    mimetypes = ["text/csv", "application/json"]


@app.post("/book")
def get_book(raw: BookRaw):
    # raw equals to flask.request
    print(raw.data)
    print(raw.mimetype)
    return "ok"


if __name__ == "__main__":
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/4/1 16:54
import json
from json import JSONDecodeError
from typing import Any, Type, Optional

from flask import request, current_app, abort
from pydantic import ValidationError, BaseModel
from pydantic.fields import FieldInfo
from werkzeug.datastructures.structures import MultiDict


def _get_list_value(model: Type[BaseModel], args: MultiDict, model_field_key: str, model_field_value: FieldInfo):
    if model_field_value.alias and model.model_config.get("populate_by_name"):
        key = model_field_value.alias
        value = args.getlist(model_field_value.alias) or args.getlist(model_field_key)
    elif model_field_value.alias:
        key = model_field_value.alias
        value = args.getlist(model_field_value.alias)
    else:
        key = model_field_key
        value = args.getlist(model_field_key)

    return key, value


def _get_value(model: Type[BaseModel], args: MultiDict, model_field_key: str, model_field_value: FieldInfo):
    if model_field_value.alias and model.model_config.get("populate_by_name"):
        key = model_field_value.alias
        value = args.get(model_field_value.alias) or args.get(model_field_key)
    elif model_field_value.alias:
        key = model_field_value.alias
        value = args.get(model_field_value.alias)
    else:
        key = model_field_key
        value = args.get(model_field_key)

    return key, value


def _validate_header(header: Type[BaseModel], func_kwargs: dict):
    request_headers = dict(request.headers)
    header_dict = {}
    model_properties = header.model_json_schema().get("properties", {})
    for model_field_key, model_field_value in header.model_fields.items():
        key_title = model_field_key.replace("_", "-").title()
        model_field_schema = model_properties.get(model_field_value.alias or model_field_key)
        if model_field_value.alias and header.model_config.get("populate_by_name"):
            key = model_field_value.alias
            key_alias_title = model_field_value.alias.replace("_", "-").title()
            value = request_headers.get(key_alias_title) or request_headers.get(key_title)
        elif model_field_value.alias:
            key = model_field_value.alias
            key_alias_title = model_field_value.alias.replace("_", "-").title()
            value = request_headers.get(key_alias_title)
        else:
            key = model_field_key
            value = request_headers[key_title]
        if value is not None:
            header_dict[key] = value
        if model_field_schema.get("type") == "null":
            header_dict[key] = value  # type:ignore
    # extra keys
    for key, value in request_headers.items():
        if key not in header_dict.keys():
            header_dict[key] = value
    func_kwargs["header"] = header.model_validate(obj=header_dict)


def _validate_cookie(cookie: Type[BaseModel], func_kwargs: dict):
    request_cookies = dict(request.cookies)
    func_kwargs["cookie"] = cookie.model_validate(obj=request_cookies)


def _validate_path(path: Type[BaseModel], path_kwargs: dict, func_kwargs: dict):
    func_kwargs["path"] = path.model_validate(obj=path_kwargs)


def _validate_query(query: Type[BaseModel], func_kwargs: dict):
    request_args = request.args
    query_dict = {}
    model_properties = query.model_json_schema().get("properties", {})
    for model_field_key, model_field_value in query.model_fields.items():
        model_field_schema = model_properties.get(model_field_value.alias or model_field_key)
        if model_field_schema.get("type") == "array":
            key, value = _get_list_value(query, request_args, model_field_key, model_field_value)
        else:
            key, value = _get_value(query, request_args, model_field_key, model_field_value)
        if value is not None and value != []:
            query_dict[key] = value
        if model_field_schema.get("type") == "null":
            query_dict[key] = value
    # extra keys
    for key, value in request_args.items():
        if key not in query_dict.keys():
            query_dict[key] = value
    func_kwargs["query"] = query.model_validate(obj=query_dict)


def _validate_form(form: Type[BaseModel], func_kwargs: dict):
    request_form = request.form
    request_files = request.files
    form_dict = {}
    model_properties = form.model_json_schema().get("properties", {})
    for model_field_key, model_field_value in form.model_fields.items():
        model_field_schema = model_properties.get(model_field_value.alias or model_field_key)
        if model_field_schema.get("type") == "array":
            if model_field_schema.get("items") == {"format": "binary", "type": "string"}:
                # list[FileStorage]
                key, value = _get_list_value(form, request_files, model_field_key, model_field_value)
            else:
                value = []
                key, value_list = _get_list_value(form, request_form, model_field_key, model_field_value)
                for _value in value_list:
                    try:
                        value.append(json.loads(_value))
                    except (JSONDecodeError, TypeError):
                        value.append(_value)
        elif model_field_schema.get("type") == "string" and model_field_schema.get("format") == "binary":
            # FileStorage
            key, value = _get_value(form, request_files, model_field_key, model_field_value)
        else:
            key, _value = _get_value(form, request_form, model_field_key, model_field_value)
            try:
                value = json.loads(_value)
            except (JSONDecodeError, TypeError):
                value = _value
        if value is not None and value != []:
            form_dict[key] = value
        if model_field_schema.get("type") == "null":
            form_dict[key] = value
    # extra keys
    for key, value in {**dict(request_form), **dict(request_files)}.items():
        if key not in form_dict.keys():
            form_dict[key] = value
    func_kwargs["form"] = form.model_validate(obj=form_dict)


def _validate_body(body: Type[BaseModel], func_kwargs: dict):
    obj = request.get_json(silent=True)
    if isinstance(obj, str):
        body_model = body.model_validate_json(json_data=obj)
    else:
        body_model = body.model_validate(obj=obj)
    func_kwargs["body"] = body_model


def _validate_request(
        header: Optional[Type[BaseModel]] = None,
        cookie: Optional[Type[BaseModel]] = None,
        path: Optional[Type[BaseModel]] = None,
        query: Optional[Type[BaseModel]] = None,
        form: Optional[Type[BaseModel]] = None,
        body: Optional[Type[BaseModel]] = None,
        raw: Optional[Type[BaseModel]] = None,
        path_kwargs: Optional[dict[Any, Any]] = None
) -> dict:
    """
    Validate requests and responses.

    Args:
        header: Header model.
        cookie: Cookie model.
        path: Path model.
        query: Query model.
        form: Form model.
        body: Body model.
        path_kwargs: Path parameters.

    Returns:
        dict: Request kwargs.

    Raises:
        ValidationError: If validation fails.
    """

    # Dictionary to store func kwargs
    func_kwargs: dict = {}

    try:
        # Validate header, cookie, path, and query parameters
        if header:
            _validate_header(header, func_kwargs)
        if cookie:
            _validate_cookie(cookie, func_kwargs)
        if path:
            _validate_path(path, path_kwargs or {}, func_kwargs)
        if query:
            _validate_query(query, func_kwargs)
        if form:
            _validate_form(form, func_kwargs)
        if body:
            _validate_body(body, func_kwargs)
        if raw:
            func_kwargs["raw"] = request
    except ValidationError as e:
        # Create a response with validation error details
        validation_error_callback = getattr(current_app, "validation_error_callback")
        abort(validation_error_callback(e))

    return func_kwargs



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:50
import typing
from typing import Optional, Union

from pydantic import BaseModel, Field

from .parameter import Parameter
from .reference import Reference
from .server import Server

if typing.TYPE_CHECKING:  # pragma: no cover
    from .operation import Operation


class PathItem(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#path-item-object
    """
    ref: Optional[str] = Field(default=None, alias="$ref")
    summary: Optional[str] = None
    description: Optional[str] = None
    get: Optional["Operation"] = None
    put: Optional["Operation"] = None
    post: Optional["Operation"] = None
    delete: Optional["Operation"] = None
    options: Optional["Operation"] = None
    head: Optional["Operation"] = None
    patch: Optional["Operation"] = None
    trace: Optional["Operation"] = None
    servers: Optional[list[Server]] = None
    parameters: Optional[list[Union[Parameter, Reference]]] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/1/1 16:54

import pytest

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI

info = Info(title='book API', version='1.0.0')

app = OpenAPI(__name__, info=info)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.get('/book', summary='new summary', description='new description')
def get_book():
    """Get a book
    to get some book by id, like:
    http://localhost:5000/book/3
    """
    return {"code": 0, "message": "ok"}  # pragma: no cover


@app.get('/book2', description='new description')
def get_book2():
    """Get a book
    to get some book by id, like:
    http://localhost:5000/book/3
    """
    return {"code": 0, "message": "ok"}  # pragma: no cover


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert resp.status_code == 200
    assert _json['paths']['/book']['get']['summary'] == 'new summary'
    assert _json['paths']['/book']['get']['description'] == 'new description'
    assert _json['paths']['/book2']['get']['summary'] == 'Get a book'
    assert _json['paths']['/book2']['get']['description'] == 'new description'



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/11 16:17
from typing import Any

from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from werkzeug.datastructures import FileStorage as _FileStorage


class FileStorage(_FileStorage):
    """
    An uploaded file included as part of the request data.
    """

    @classmethod
    def __get_pydantic_json_schema__(cls, *_args: Any, **_kwargs: Any) -> JsonSchemaValue:
        field_schema = {"format": "binary", "type": "string"}
        return field_schema

    @classmethod
    def __get_pydantic_core_schema__(cls, *_args: Any, **_kwargs: Any) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, value: _FileStorage, *_args: Any, **_kwargs: Any) -> _FileStorage:
        return value



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/9/7 9:46
import pytest
from flask import request

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)


@app.get("/t")
@app.get("/tt")
def get_t():
    return request.url


@app.post("/t")
@app.post("/tt")
def post_t():
    return request.url


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_get_t(client):
    r = client.get("/t")
    assert r.text == "http://localhost/t"
    r = client.get("/tt")
    assert r.text == "http://localhost/tt"


def test_post_t(client):
    r = client.post("/t")
    assert r.text == "http://localhost/t"
    r = client.post("/tt")
    assert r.text == "http://localhost/tt"



from flask_openapi3 import OpenAPI, APIBlueprint

app = OpenAPI(__name__)

api = APIBlueprint('book', __name__, url_prefix='/api/book')
api_english = APIBlueprint('english', __name__)
api_chinese = APIBlueprint('chinese', __name__)


@api_english.post('/english')
def create_english_book():
    return {"message": "english"}


@api_chinese.post('/chinese')
def create_chinese_book():
    return {"message": "chinese"}


# register nested api
api.register_api(api_english)
api.register_api(api_chinese)
# register api
app.register_api(api)

if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:42
from typing import Any, Optional

from pydantic import BaseModel


class Example(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#example-object
    """

    summary: Optional[str] = None
    description: Optional[str] = None
    value: Optional[Any] = None
    externalValue: Optional[str] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:38
from enum import Enum


class DataType(str, Enum):
    """
    https://spec.openapis.org/oas/v3.1.0#data-types
    """

    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"
    NUll = "null"



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/6 16:38
from __future__ import annotations

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI

info = Info(title='book API', version='1.0.0')

jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
security_schemes = {"jwt": jwt}


class NotFoundResponse(BaseModel):
    code: int = Field(-1, description="Status Code")
    message: str = Field("Resource not found!", description="Exception Information")


doc_prefix = '/v1/openapi'
app = OpenAPI(
    __name__,
    info=info,
    doc_prefix=doc_prefix,
    security_schemes=security_schemes,
    responses={"404": NotFoundResponse}
)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_openapi(client):
    resp = client.get(f"{doc_prefix}/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/11/10 12:17
from pydantic import ValidationError

from flask_openapi3 import Server, ServerVariable


def test_server_variable():
    Server(
        url="http://127.0.0.1:5000",
        variables=None
    )
    try:
        variables = {"one": ServerVariable(default="one", enum=[])}
        Server(
            url="http://127.0.0.1:5000",
            variables=variables
        )
        error = 0
    except ValidationError:
        error = 1
    assert error == 1
    try:
        variables = {"one": ServerVariable(default="one")}
        Server(
            url="http://127.0.0.1:5000",
            variables=variables
        )
        error = 0
    except ValidationError:
        error = 1
    assert error == 0
    try:
        variables = {"one": ServerVariable(default="one", enum=["one", "two"])}
        Server(
            url="http://127.0.0.1:5000",
            variables=variables
        )
        error = 0
    except ValidationError:
        error = 1
    assert error == 0



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:53
from pydantic import BaseModel, Field


class Reference(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#reference-object
    """

    ref: str = Field(..., alias="$ref")

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/1 21:34

import inspect
import re
import sys
from enum import Enum
from http import HTTPStatus
from typing import get_type_hints, Type, Callable, Optional, Any, DefaultDict

from flask import make_response, current_app
from flask.wrappers import Response as FlaskResponse
from pydantic import BaseModel, ValidationError
from pydantic.json_schema import JsonSchemaMode

from .models import Encoding
from .models import MediaType
from .models import OPENAPI3_REF_PREFIX
from .models import OPENAPI3_REF_TEMPLATE
from .models import Operation
from .models import Parameter
from .models import ParameterInType
from .models import PathItem
from .models import RawModel
from .models import RequestBody
from .models import Response
from .models import Schema
from .models import Tag
from .models.data_type import DataType
from .types import ParametersTuple
from .types import ResponseDict
from .types import ResponseStrKeyDict

HTTP_STATUS = {str(status.value): status.phrase for status in HTTPStatus}

if sys.version_info < (3, 11):  # pragma: no cover

    class HTTPMethod(str, Enum):
        GET = "GET"
        POST = "POST"
        PUT = "PUT"
        DELETE = "DELETE"
        PATCH = "PATCH"
        HEAD = "HEAD"
        OPTIONS = "OPTIONS"
        TRACE = "TRACE"
        CONNECT = "CONNECT"
else:
    from http import HTTPMethod


def get_operation(
        func: Callable, *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        openapi_extensions: Optional[dict[str, Any]] = None,
) -> Operation:
    """
    Return an Operation object with the specified summary and description.

    Args:
        func: The function or method for which the operation is being defined.
        summary: A short summary of what the operation does.
        description: A verbose explanation of the operation behavior.
        openapi_extensions: Additional extensions to the OpenAPI Schema.

    Returns:
        An Operation object representing the operation.

    """
    # Get the docstring of the function
    doc = inspect.getdoc(func) or ""
    doc = doc.strip()
    lines = doc.split("\n")
    doc_summary = lines[0]

    # Determine the summary and description based on provided arguments or docstring
    if summary is None:
        doc_description = lines[0] if len(lines) == 0 else "</br>".join(lines[1:])
    else:
        doc_description = "</br>".join(lines)

    summary = summary or doc_summary
    description = description or doc_description

    # Create the operation dictionary with summary and description
    operation_dict = {}

    if summary:
        operation_dict["summary"] = summary  # type: ignore

    if description:
        operation_dict["description"] = description  # type: ignore

    # Add any additional openapi_extensions to the operation dictionary
    operation_dict.update(openapi_extensions or {})

    # Create and return the Operation object
    operation = Operation(**operation_dict)

    return operation


def get_operation_id_for_path(*, name: str, path: str, method: str) -> str:
    """
    Generate a unique operation ID based on the name, path, and method.

    Args:
        name: The name or identifier for the operation.
        path: The URL path for the operation.
        method: The HTTP method for the operation.

    Returns:
        A unique operation ID generated based on the provided name, path, and method.

    """

    return re.sub(r"\W", "_", name + path) + "_" + method.lower()


def get_model_schema(model: Type[BaseModel], mode: JsonSchemaMode = "validation") -> dict:
    """Converts a Pydantic model to an OpenAPI schema."""

    assert inspect.isclass(model) and issubclass(model, BaseModel), \
        f"{model} is invalid `pydantic.BaseModel`"

    model_config = model.model_config
    by_alias = bool(model_config.get("by_alias", True))

    return model.model_json_schema(by_alias=by_alias, ref_template=OPENAPI3_REF_TEMPLATE, mode=mode)


def parse_header(header: Type[BaseModel]) -> tuple[list[Parameter], dict]:
    """Parses a header model and returns a list of parameters and component schemas."""
    schema = get_model_schema(header)
    parameters = []
    components_schemas: dict = dict()
    properties = schema.get("properties", {})

    for name, value in properties.items():
        data = {
            "name": name,
            "in": ParameterInType.HEADER,
            "required": name in schema.get("required", []),
            "schema": Schema(**value)
        }
        # Parse extra values
        if "description" in value.keys():
            data["description"] = value.get("description")
        if "deprecated" in value.keys():
            data["deprecated"] = value.get("deprecated")
        if "example" in value.keys():
            data["example"] = value.get("example")
        if "examples" in value.keys():
            data["examples"] = value.get("examples")
        parameters.append(Parameter(**data))

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return parameters, components_schemas


def parse_cookie(cookie: Type[BaseModel]) -> tuple[list[Parameter], dict]:
    """Parses a cookie model and returns a list of parameters and component schemas."""
    schema = get_model_schema(cookie)
    parameters = []
    components_schemas: dict = dict()
    properties = schema.get("properties", {})

    for name, value in properties.items():
        data = {
            "name": name,
            "in": ParameterInType.COOKIE,
            "required": name in schema.get("required", []),
            "schema": Schema(**value)
        }
        # Parse extra values
        if "description" in value.keys():
            data["description"] = value.get("description")
        if "deprecated" in value.keys():
            data["deprecated"] = value.get("deprecated")
        if "example" in value.keys():
            data["example"] = value.get("example")
        if "examples" in value.keys():
            data["examples"] = value.get("examples")
        parameters.append(Parameter(**data))

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return parameters, components_schemas


def parse_path(path: Type[BaseModel]) -> tuple[list[Parameter], dict]:
    """Parses a path model and returns a list of parameters and component schemas."""
    schema = get_model_schema(path)
    parameters = []
    components_schemas: dict = dict()
    properties = schema.get("properties", {})

    for name, value in properties.items():
        data = {
            "name": name,
            "in": ParameterInType.PATH,
            "required": True,
            "schema": Schema(**value)
        }
        # Parse extra values
        if "description" in value.keys():
            data["description"] = value.get("description")
        if "deprecated" in value.keys():
            data["deprecated"] = value.get("deprecated")
        if "example" in value.keys():
            data["example"] = value.get("example")
        if "examples" in value.keys():
            data["examples"] = value.get("examples")
        parameters.append(Parameter(**data))

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return parameters, components_schemas


def parse_query(query: Type[BaseModel]) -> tuple[list[Parameter], dict]:
    """Parses a query model and returns a list of parameters and component schemas."""
    schema = get_model_schema(query)
    parameters = []
    components_schemas: dict = dict()
    properties = schema.get("properties", {})

    for name, value in properties.items():
        data = {
            "name": name,
            "in": ParameterInType.QUERY,
            "required": name in schema.get("required", []),
            "schema": Schema(**value)
        }
        # Parse extra values
        if "description" in value.keys():
            data["description"] = value.get("description")
        if "deprecated" in value.keys():
            data["deprecated"] = value.get("deprecated")
        if "example" in value.keys():
            data["example"] = value.get("example")
        if "examples" in value.keys():
            data["examples"] = value.get("examples")
        parameters.append(Parameter(**data))

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return parameters, components_schemas


def parse_form(
        form: Type[BaseModel],
) -> tuple[dict[str, MediaType], dict]:
    """Parses a form model and returns a list of parameters and component schemas."""
    schema = get_model_schema(form)
    components_schemas = dict()
    properties = schema.get("properties", {})

    assert properties, f"{form.__name__}'s properties cannot be empty."

    original_title = schema.get("title") or form.__name__
    title = normalize_name(original_title)
    components_schemas[title] = Schema(**schema)
    encoding = {}
    for k, v in properties.items():
        if v.get("type") == "array":
            encoding[k] = Encoding(style="form", explode=True)
    content = {
        "multipart/form-data": MediaType(
            schema=Schema(**{"$ref": f"{OPENAPI3_REF_PREFIX}/{title}"}),
        )
    }
    if encoding:
        content["multipart/form-data"].encoding = encoding

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return content, components_schemas


def parse_body(
        body: Type[BaseModel],
) -> tuple[dict[str, MediaType], dict]:
    """Parses a body model and returns a list of parameters and component schemas."""
    schema = get_model_schema(body)
    components_schemas = dict()

    original_title = schema.get("title") or body.__name__
    title = normalize_name(original_title)
    components_schemas[title] = Schema(**schema)
    content = {
        "application/json": MediaType(
            schema=Schema(**{"$ref": f"{OPENAPI3_REF_PREFIX}/{title}"})
        )
    }

    # Parse definitions
    definitions = schema.get("$defs", {})
    for name, value in definitions.items():
        components_schemas[name] = Schema(**value)

    return content, components_schemas


def get_responses(
        responses: ResponseStrKeyDict,
        components_schemas: dict,
        operation: Operation
) -> None:
    _responses = {}
    _schemas = {}

    for key, response in responses.items():
        if response is None:
            # If the response is None, it means HTTP status code "204" (No Content)
            _responses[key] = Response(description=HTTP_STATUS.get(key, ""))
        elif isinstance(response, dict):
            response["description"] = response.get("description", HTTP_STATUS.get(key, ""))
            _responses[key] = Response(**response)
        else:
            # OpenAPI 3 support ^[a-zA-Z0-9\.\-_]+$ so we should normalize __name__
            schema = get_model_schema(response, mode="serialization")
            original_title = schema.get("title") or response.__name__
            name = normalize_name(original_title)
            _responses[key] = Response(
                description=HTTP_STATUS.get(key, ""),
                content={
                    "application/json": MediaType(
                        schema=Schema(**{"$ref": f"{OPENAPI3_REF_PREFIX}/{name}"})
                    )})

            model_config: DefaultDict[str, Any] = response.model_config  # type: ignore
            openapi_extra = model_config.get("openapi_extra", {})
            if openapi_extra:
                openapi_extra_keys = openapi_extra.keys()
                # Add additional information from model_config to the response
                if "description" in openapi_extra_keys:
                    _responses[key].description = openapi_extra.get("description")
                if "headers" in openapi_extra_keys:
                    _responses[key].headers = openapi_extra.get("headers")
                if "links" in openapi_extra_keys:
                    _responses[key].links = openapi_extra.get("links")
                _content = _responses[key].content
                if "example" in openapi_extra_keys:
                    _content["application/json"].example = openapi_extra.get("example")  # type: ignore
                if "examples" in openapi_extra_keys:
                    _content["application/json"].examples = openapi_extra.get("examples")  # type: ignore
                if "encoding" in openapi_extra_keys:
                    _content["application/json"].encoding = openapi_extra.get("encoding")  # type: ignore
                _content.update(openapi_extra.get("content", {}))  # type: ignore

            _schemas[name] = Schema(**schema)
            definitions = schema.get("$defs")
            if definitions:
                # Add schema definitions to _schemas
                for name, value in definitions.items():
                    _schemas[normalize_name(name)] = Schema(**value)

    components_schemas.update(**_schemas)
    operation.responses = _responses


def parse_and_store_tags(
        new_tags: list[Tag],
        old_tags: list[Tag],
        old_tag_names: list[str],
        operation: Operation
) -> None:
    """
    Parses new tags, stores them in an old_tags list if they are not already present,
    and updates the tags attribute of the operation object.

    Args:
        new_tags: A list of new Tag objects to be parsed and stored.
        old_tags: The list of existing Tag objects.
        old_tag_names: The list that names of existing tags.
        operation: The operation object whose tag attribute needs to be updated.

    Returns:
        None
    """
    # Iterate over each tag in new_tags
    for tag in new_tags:
        if tag.name not in old_tag_names:
            old_tag_names.append(tag.name)
            old_tags.append(tag)

    # Set the tags attribute of the operation object to a list of unique tag names from new_tags
    # If the resulting list is empty, set it to ["default"]
    operation.tags = list(set([tag.name for tag in new_tags])) or ["default"]


def parse_parameters(
        func: Callable,
        *,
        components_schemas: Optional[dict] = None,
        operation: Optional[Operation] = None,
        doc_ui: bool = True,
) -> ParametersTuple:
    """
    Parses the parameters of a given function and returns the types for header, cookie, path,
    query, form, and body parameters. Also populates the Operation object with the parsed parameters.

    Args:
        func: The function to parse the parameters from.
        components_schemas: Dictionary to store the parsed components schemas (default: None).
        operation: Operation object to populate with parsed parameters (default: None).
        doc_ui: Flag indicating whether to return types for documentation UI (default: True).

    Returns:
        tuple[Type[BaseModel], Type[BaseModel], Type[BaseModel], Type[BaseModel], Type[BaseModel], Type[BaseModel]]:
        The types for header, cookie, path, query, form, and body parameters respectively.

    """

    # If components_schemas is None, initialize it as an empty dictionary
    if components_schemas is None:
        components_schemas = dict()

    # If operation is None, initialize it as an Operation object
    if operation is None:
        operation = Operation()

    # Get the type hints from the function
    annotations = get_type_hints(func)

    # Get the types for header, cookie, path, query, form, and body parameters
    header: Optional[Type[BaseModel]] = annotations.get("header")
    cookie: Optional[Type[BaseModel]] = annotations.get("cookie")
    path: Optional[Type[BaseModel]] = annotations.get("path")
    query: Optional[Type[BaseModel]] = annotations.get("query")
    form: Optional[Type[BaseModel]] = annotations.get("form")
    body: Optional[Type[BaseModel]] = annotations.get("body")
    raw: Optional[Type[RawModel]] = annotations.get("raw")

    # If doc_ui is False, return the types without further processing
    if doc_ui is False:
        return header, cookie, path, query, form, body, raw

    parameters = []

    if header:
        _parameters, _components_schemas = parse_header(header)
        parameters.extend(_parameters)
        components_schemas.update(**_components_schemas)

    if cookie:
        _parameters, _components_schemas = parse_cookie(cookie)
        parameters.extend(_parameters)
        components_schemas.update(**_components_schemas)

    if path:
        _parameters, _components_schemas = parse_path(path)
        parameters.extend(_parameters)
        components_schemas.update(**_components_schemas)

    if query:
        _parameters, _components_schemas = parse_query(query)
        parameters.extend(_parameters)
        components_schemas.update(**_components_schemas)

    if form:
        _content, _components_schemas = parse_form(form)
        components_schemas.update(**_components_schemas)
        request_body = RequestBody(content=_content, required=True)
        model_config: DefaultDict[str, Any] = form.model_config  # type: ignore
        openapi_extra = model_config.get("openapi_extra", {})
        if openapi_extra:
            openapi_extra_keys = openapi_extra.keys()
            if "description" in openapi_extra_keys:
                request_body.description = openapi_extra.get("description")
            if "example" in openapi_extra_keys:
                request_body.content["multipart/form-data"].example = openapi_extra.get("example")
            if "examples" in openapi_extra_keys:
                request_body.content["multipart/form-data"].examples = openapi_extra.get("examples")
            if "encoding" in openapi_extra_keys:
                request_body.content["multipart/form-data"].encoding = openapi_extra.get("encoding")
        operation.requestBody = request_body

    if body:
        _content, _components_schemas = parse_body(body)
        components_schemas.update(**_components_schemas)
        request_body = RequestBody(content=_content, required=True)
        model_config: DefaultDict[str, Any] = body.model_config  # type: ignore
        openapi_extra = model_config.get("openapi_extra", {})
        if openapi_extra:
            openapi_extra_keys = openapi_extra.keys()
            if "description" in openapi_extra_keys:
                request_body.description = openapi_extra.get("description")
            request_body.required = openapi_extra.get("required", True)
            if "example" in openapi_extra_keys:
                request_body.content["application/json"].example = openapi_extra.get("example")
            if "examples" in openapi_extra_keys:
                request_body.content["application/json"].examples = openapi_extra.get("examples")
            if "encoding" in openapi_extra_keys:
                request_body.content["application/json"].encoding = openapi_extra.get("encoding")
        operation.requestBody = request_body

    if raw:
        _content = {}
        for mimetype in raw.mimetypes:
            if mimetype.startswith("application/json"):
                _content[mimetype] = MediaType(
                    schema=Schema(type=DataType.OBJECT)
                )
            else:
                _content[mimetype] = MediaType(
                    schema=Schema(type=DataType.STRING)
                )
        request_body = RequestBody(content=_content)
        operation.requestBody = request_body

    if parameters:
        # Set the parsed parameters in the operation object
        operation.parameters = parameters

    return header, cookie, path, query, form, body, raw


def parse_method(uri: str, method: str, paths: dict, operation: Operation) -> None:
    """
    Parses the HTTP method and updates the corresponding PathItem object in the paths' dictionary.

    Args:
        uri: The URI of the API endpoint.
        method: The HTTP method for the API endpoint.
        paths: A dictionary containing the API paths and their corresponding PathItem objects.
        operation: The Operation object to assign to the PathItem.

    Returns:
        None
    """
    # Check the HTTP method and update the PathItem object in the path dictionary
    if method == HTTPMethod.GET:
        if not paths.get(uri):
            paths[uri] = PathItem(get=operation)
        else:
            paths[uri].get = operation
    elif method == HTTPMethod.POST:
        if not paths.get(uri):
            paths[uri] = PathItem(post=operation)
        else:
            paths[uri].post = operation
    elif method == HTTPMethod.PUT:
        if not paths.get(uri):
            paths[uri] = PathItem(put=operation)
        else:
            paths[uri].put = operation
    elif method == HTTPMethod.PATCH:
        if not paths.get(uri):
            paths[uri] = PathItem(patch=operation)
        else:
            paths[uri].patch = operation
    elif method == HTTPMethod.DELETE:
        if not paths.get(uri):
            paths[uri] = PathItem(delete=operation)
        else:
            paths[uri].delete = operation


def make_validation_error_response(e: ValidationError) -> FlaskResponse:
    """
    Create a Flask response for a validation error.

    Args:
        e: The ValidationError object containing the details of the error.

    Returns:
        FlaskResponse: A Flask Response object with the JSON representation of the error.
    """
    response = make_response(e.json())
    response.headers["Content-Type"] = "application/json"
    response.status_code = getattr(current_app, "validation_error_status", 422)
    return response


def parse_rule(rule: str, url_prefix=None) -> str:
    trail_slash = rule.endswith("/")

    # Merge url_prefix and uri
    uri = url_prefix.rstrip("/") + "/" + rule.lstrip("/") if url_prefix else rule

    if not trail_slash:
        uri = uri.rstrip("/")

    # Convert a route parameter format from /pet/<petId> to /pet/{petId}
    uri = re.sub(r"<([^<:]+:)?", "{", uri).replace(">", "}")

    return uri


def convert_responses_key_to_string(responses: ResponseDict) -> ResponseStrKeyDict:
    """Convert key to string"""

    return {str(key.value if isinstance(key, HTTPStatus) else key): value for key, value in responses.items()}


def normalize_name(name: str) -> str:
    return re.sub(r"[^\w.\-]", "_", name)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/8/18 17:19
from enum import Enum

from pydantic import BaseModel, Field

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI

app = OpenAPI(
    __name__,
    info=Info(title='Enum demo', version='1.0.0')
)


class Language(str, Enum):
    cn = 'Chinese'
    en = 'English'


class LanguagePath(BaseModel):
    language: Language = Field(..., description='Language')


@app.get('/<language>')
def get_enum(path: LanguagePath):
    print(path)
    return {}


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/12/19 10:34

from flask_openapi3.utils import normalize_name


def test_normalize_name():
    assert "List-Generic.Response_Detail_" == normalize_name("List-Generic.Response[Detail]")



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/4/16 14:21
import json

from flask import current_app
from flask.cli import click
from flask.cli import with_appcontext


@click.command(name="openapi")
@click.option("--output", "-o", type=click.Path(), help="The output file path.")
@click.option("--format", "-f", "_format", type=click.Choice(["json", "yaml"]), help="The output file format.")
@click.option("--indent", "-i", type=int, help="The indentation for JSON dumps.")
@with_appcontext
def openapi_command(output, _format, indent):
    """Export the OpenAPI Specification to console or a file"""

    # Check if the current app has an api_doc attribute
    if hasattr(current_app, "api_doc"):
        obj = current_app.api_doc

        # Generate the OpenAPI Specification based on the specified format
        if _format == "yaml":
            try:
                import yaml  # type: ignore
            except ImportError:  # pragma: no cover
                raise ImportError("pyyaml must be installed.")
            openapi = yaml.safe_dump(obj, allow_unicode=True)
        else:
            openapi = json.dumps(obj, indent=indent, ensure_ascii=False)

        # Save the OpenAPI Specification to a file if the output path is provided
        if output:
            with open(output, "w", encoding="utf8") as f:
                f.write(openapi)
            click.echo(f"Saved to {output}.")
        else:
            click.echo(openapi)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/8/31 15:35
from typing import Sequence

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


class BookQuery(BaseModel):
    age: int
    author: str = Field(..., alias="author_name")

    model_config = {"populate_by_name": True}


@app.get("/book", summary="get books")
def get_book(query: BookQuery):
    """
    get all books
    """
    print(query)
    return "ok"


class QueryModel(BaseModel):
    aliased_field: str = Field(alias="aliasedField")

    aliased_list_field: list[str] = Field(alias="aliasedListField")


@app.get("/query-alias-test")
def query_alias_test(query: QueryModel):
    assert query.aliased_field == "test"
    return "ok"


class HeaderModel(BaseModel):
    Hello1: str = Field(..., alias="Hello2")

    model_config = {"populate_by_name": True}


@app.get("/header")
def get_book_header(header: HeaderModel):
    return header.model_dump(by_alias=True)


class TupleModel(BaseModel):
    values: tuple[int, int]
    sequence: Sequence[int] = Field(alias="Sequence")

    model_config = {"populate_by_name": True}


@app.get("/tuple-test")
def tuple_test(query: TupleModel):
    assert query.values == (2, 2)
    return b"", 200


class AliasModel(BaseModel):
    aliased_field: str = Field(alias="aliasedField")


@app.post("/form-alias-test")
def alias_test(form: AliasModel):
    assert form.aliased_field == "test"
    return b"", 200


def test_header(client):
    headers = {"Hello2": "111"}
    resp = client.get("/header", headers=headers)
    print(resp.json)
    assert resp.status_code == 200
    assert resp.json == headers


def test_tuple_query(client):
    resp = client.get(
        "/tuple-test",
        query_string={"values": [2, 2], "sequence": [1, 2, 3]},
    )
    assert resp.status_code == 200


def test_form_alias(client):
    resp = client.post(
        "/form-alias-test",
        data={"aliasedField": "test"},
    )
    assert resp.status_code == 200

    resp = client.post(
        "/form-alias-test",
        data={"aliased_field": "test"},
    )
    assert resp.status_code == 422


def test_query_alias(client):
    resp = client.get(
        "/query-alias-test",
        query_string={"aliasedField": "test", "aliasedListField": ["test"]},
    )
    assert resp.status_code == 200

    resp = client.get(
        "/query-alias-test",
        data={"aliased_field": "test", "aliased_list_field": ["test"]},
    )
    assert resp.status_code == 422


def test_query_populate_by_name(client):
    resp = client.get("/book?age=1&author=aa")
    assert resp.status_code == 200
    resp = client.get("/book?age=1&author_name=aa")
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/9 11:23
from http import HTTPStatus

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, APIBlueprint

app = OpenAPI(__name__)
app.config["TESTING"] = True
api = APIBlueprint("/api", __name__, url_prefix="/api")


class BookResponse(BaseModel):
    code: int = Field(0, description="Status Code")
    message: str = Field("ok", description="Exception Information")


class BookPath(BaseModel):
    bid: int = Field(..., description="book id")


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.get(
    "/book/<int:bid>",
    responses={
        HTTPStatus.OK: BookResponse,
        "201": BookResponse,
        202: {"content": {"text/html": {"schema": {"type": "string"}}}},
        204: None,
        "422": {"description": "validation error"}
    }
)
def get_book(path: BookPath):
    print(path)  # pragma: no cover


@api.get("/book", responses={
    HTTPStatus.OK: BookResponse,
    "201": BookResponse, 204: None
})
def get_api_book():
    return {"code": 0, "message": "ok"}  # pragma: no cover


app.register_api(api)


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert resp.status_code == 200
    assert _json["paths"]["/book/{bid}"]["get"]["responses"].keys() - ["200", "201", "202", "204"] == {"422"}
    assert _json["paths"]["/api/book"]["get"]["responses"].keys() - ["200", "201", "202", "204"] == set()



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/10 14:51
from typing import Any, Optional

from pydantic import BaseModel, Field


class ValidationErrorModel(BaseModel):
    # More information: https://docs.pydantic.dev/latest/usage/models/#error-handling
    loc: Optional[list[str]] = Field(None, title="Location", description="the error's location as a list. ")
    msg: Optional[str] = Field(None, title="Message", description="a computer-readable identifier of the error type.")
    type_: Optional[str] = Field(None, title="Error Type", description="a human readable explanation of the error.")
    ctx: Optional[dict[str, Any]] = Field(
        None,
        title="Error context",
        description="an optional object which contains values required to render the error message."
    )


# backward compatibility
UnprocessableEntity = ValidationErrorModel



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:45
from typing import Any, Optional

from pydantic import BaseModel

from .server import Server


class Link(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#link-object
    """

    operationRef: Optional[str] = None
    operationId: Optional[str] = None
    parameters: Optional[dict[str, Any]] = None
    requestBody: Optional[Any] = None
    description: Optional[str] = None
    server: Optional[Server] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/11/30 14:55

from typing import Optional

from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, APIView

app = OpenAPI(__name__)
api_view = APIView(url_prefix="/api/v1")


class Query(BaseModel):
    q: str


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description='Age')


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


@app.get('/open/api')
async def get_openapi(query: Query):
    print(query)
    return 'GET, OpenAPI!'


@app.post('/open/api')
async def post_openapi(body: Query):
    print(body)
    return 'POST, OpenAPI!'


@api_view.route("/book")
class BookListAPIView:

    @api_view.doc(summary="get book list")
    async def get(self, query: BookQuery):
        return query.model_dump_json()

    @api_view.doc(summary="create book")
    async def post(self, body: BookBody):
        """description for a created book"""
        return body.model_dump_json()


app.register_api_view(api_view)

if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:49
from typing import Any, Optional, Union

from pydantic import BaseModel, Field

from .example import Example
from .media_type import MediaType
from .parameter_in_type import ParameterInType
from .reference import Reference
from .schema import Schema


class Parameter(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#parameter-object
    """

    name: str
    param_in: ParameterInType = Field(alias="in")
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: Optional[bool] = None
    allowEmptyValue: Optional[bool] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: Optional[bool] = None
    param_schema: Optional[Union[Reference, Schema]] = Field(default=None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[dict[str, Union[Example, Reference]]] = None
    content: Optional[dict[str, MediaType]] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/12/9 17:50

import orjson
from flask.json.provider import JSONProvider

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI


class OrJSONProvider(JSONProvider):
    # https://github.com/ijl/orjson#option
    option = orjson.OPT_INDENT_2

    def dumps(self, obj, **kwargs):
        return orjson.dumps(obj, option=self.option).decode()

    def loads(self, s, **kwargs):
        return orjson.loads(s)


info = Info(title='book API', version='1.0.0')
app = OpenAPI(__name__, info=info)
# use orjson
orjson_provider = OrJSONProvider(app)
app.json = orjson_provider


@app.get('/book')
def get_book():
    return {
        "code": 0,
        "message": "ok",
        "data": [
            {"bid": 1, "age": 18, "author": "tom"},
            {"bid": 2, "age": 19, "author": "alice"},
            {"bid": 3, "age": 20, "author": "中文测试"}
        ]
    }


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/11/20 14:45
from typing import Optional

import pytest
from pydantic import BaseModel, Field, ConfigDict

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description="Age")

    model_config = ConfigDict(extra="allow")


class BookForm(BaseModel):
    string: str

    model_config = ConfigDict(extra="forbid")


class BookHeader(BaseModel):
    api_key: str = Field(..., description="API Key")

    model_config = ConfigDict(extra="forbid")


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.get("/book")
def get_books(query: BookQuery):
    """get books
    to get all books
    """
    assert query.age == 3
    assert query.author == "joy"
    return {"code": 0, "message": "ok"}


@app.post("/form")
def api_form(form: BookForm):
    print(form)
    return {"code": 0, "message": "ok"}


def test_query(client):
    resp = client.get("/book?age=3&author=joy")
    assert resp.status_code == 200


@app.get("/header")
def get_book(header: BookHeader):
    return header.model_dump(by_alias=True)


def test_form(client):
    data = {
        "string": "a",
        "string_list": ["a", "b", "c"]
    }
    r = client.post("/form", data=data, content_type="multipart/form-data")
    assert r.status_code == 422


def test_header(client):
    headers = {"Hello1": "111", "hello2": "222", "api_key": "333", "api_type": "A", "x-hello": "444"}
    resp = client.get("/header", headers=headers)
    assert resp.status_code == 422



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/4/19 20:53

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


class MyModel(BaseModel):
    num_1: int = Field(..., ge=1, le=10)
    num_2: int = Field(..., gt=1, lt=10)


@app.post('/book')
def create_book(body: MyModel):
    print(body)  # pragma: no cover


@pytest.fixture
def client():
    client = app.test_client()
    return client


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc

    model_props = resp.json['components']['schemas']['MyModel']['properties']
    num_1_props = model_props['num_1']
    num_2_props = model_props['num_2']

    assert num_1_props['minimum'] == 1
    assert num_1_props['maximum'] == 10
    assert num_2_props['exclusiveMinimum'] == 1
    assert num_2_props['exclusiveMaximum'] == 10



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/5/5 13:28

import pytest

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI, APIBlueprint

info = Info(title='book API', version='1.0.0')

app = OpenAPI(__name__, info=info)
app.config["TESTING"] = True

api1 = APIBlueprint('book1', __name__, url_prefix='/api/v1/book1')
api2 = APIBlueprint('book2', __name__, url_prefix='/api/v1/book2')


@pytest.fixture
def client():
    client = app.test_client()

    return client


@api1.get('/')
def get_book():
    return 'with slash'


@api2.get('')
def get_book2():
    return 'without slash'


app.register_api(api1)
app.register_api(api2)


def test_with_slash1(client):
    resp = client.get("/api/v1/book1/")
    assert resp.status_code == 200


def test_with_slash2(client):
    resp = client.get("/api/v1/book1")
    assert resp.status_code == 308


def test_without_slash1(client):
    resp = client.get("/api/v1/book2/")
    assert resp.status_code == 404


def test_without_slash2(client):
    resp = client.get("/api/v1/book2")
    assert resp.status_code == 200


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert _json['paths'].get('/api/v1/book1/') is not None
    assert _json['paths'].get('/api/v1/book2') is not None



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/5/15 14:19

import pytest

from flask_openapi3 import APIBlueprint, OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


def get_operation_id_for_path_callback(*, name: str, path: str, method: str) -> str:
    print(name, path, method)
    return name


api = APIBlueprint(
    '/book',
    __name__,
    url_prefix='/api',
    operation_id_callback=get_operation_id_for_path_callback,
)


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.get('/book', endpoint='endpoint_get_book')
def get_book():
    return 'app_book'


@api.post('/book', endpoint='endpoint_post_book')
def create_book():
    return 'api_book'


# register api
app.register_api(api)


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc
    assert resp.json["paths"]["/book"]["get"]["operationId"] == "get_book_book_get"  # Default operation_id generator
    assert resp.json["paths"]["/api/book"]["post"]["operationId"] == "/book"  # Custom callback operation_id


def test_get(client):
    resp = client.get("/book")

    assert resp.text == 'app_book'
    assert 'endpoint_get_book' in app.view_functions.keys()


def test_post(client):
    resp = client.post("/api/book")

    assert resp.text == 'api_book'
    assert '/book.endpoint_post_book' in app.view_functions.keys()



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/1/8 14:25
from pydantic import BaseModel

from flask_openapi3 import Info, Tag
from flask_openapi3 import OpenAPI, Server

info = Info(title="book API", version="1.0.0")
servers = [
    Server(url="http://127.0.0.1:5000"),
    Server(url="https://127.0.0.1:5000"),
]
app = OpenAPI(__name__, info=info, servers=servers)

book_tag = Tag(name="book", description="Some Book")


class BookQuery(BaseModel):
    age: int
    author: str


@app.get("/book", tags=[book_tag])
def get_book(query: BookQuery):
    """get books
    to get all books
    """
    return {
        "code": 0,
        "message": "ok",
        "data": [
            {"bid": 1, "age": query.age, "author": query.author},
            {"bid": 2, "age": query.age, "author": query.author}
        ]
    }


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:35
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .path_item import PathItem
else:
    PathItem = "PathItem"

"""
A map of possible out-of band callbacks related to the parent operation.
Each value in the map is a [Path Item Object](#pathItemObject)
that describes a set of requests that may be initiated by the API provider and the expected responses.
The key value used to identify the path item object is an expression, evaluated at runtime,
that identifies a URL to use for the callback operation.
"""
Callback = dict[str, PathItem]



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/11/4 14:41

from typing import Optional

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import APIView
from flask_openapi3 import OpenAPI, Tag, Info

info = Info(title='book API', version='1.0.0')
jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
security_schemes = {"jwt": jwt}

app = OpenAPI(__name__, info=info, security_schemes=security_schemes)
app.config["TESTING"] = True
security = [{"jwt": []}]

api_view = APIView(url_prefix="/api/v1/<name>", view_tags=[Tag(name="book")], view_security=security)
api_view2 = APIView(doc_ui=False)


class BookPath(BaseModel):
    id: int = Field(..., description="book ID")
    name: str


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description='Age')


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


@api_view.route("/book")
class BookListAPIView:
    a = 1

    @api_view.doc(
        summary="get book list",
        responses={
            204: None
        },
        doc_ui=False
    )
    def get(self, query: BookQuery):
        print(self.a)
        return query.model_dump_json()

    @api_view.doc(summary="create book")
    def post(self, body: BookBody):
        """description for a created book"""
        return body.model_dump_json()


@api_view.route("/book/<id>")
class BookAPIView:
    @api_view.doc(summary="get book")
    def get(self, path: BookPath):
        print(path)
        return "get"

    @api_view.doc(summary="update book", operation_id="update")
    def put(self, path: BookPath):
        print(path)
        return "put"

    @api_view.doc(summary="delete book", deprecated=True)
    def delete(self, path: BookPath):
        print(path)
        return "delete"


@api_view2.route("/<name>/book2/<id>")
class BookAPIView2:
    @api_view2.doc(summary="get book")
    def get(self, path: BookPath):
        return path.model_dump()


app.register_api_view(api_view)
app.register_api_view(api_view2)


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc
    assert resp.json["paths"]["/api/v1/{name}/book/{id}"]["put"]["operationId"] == "update"
    assert resp.json["paths"]["/api/v1/{name}/book/{id}"]["delete"][
               "operationId"] == "BookAPIView_delete_book__id__delete"


def test_get_list(client):
    resp = client.get("/api/v1/name1/book")
    assert resp.status_code == 200


def test_post(client):
    resp = client.post("/api/v1/name1/book", json={"age": 3})
    assert resp.status_code == 200


def test_put(client):
    resp = client.put("/api/v1/name1/book/1", json={"age": 3})
    assert resp.status_code == 200


def test_get(client):
    resp = client.get("/api/v1/name1/book/1")
    assert resp.status_code == 200

    resp = client.get("/name2/book2/1")
    assert resp.status_code == 200


def test_delete(client):
    resp = client.delete("/api/v1/name1/book/1")
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:44
from typing import Optional

from .parameter import Parameter
from .parameter_in_type import ParameterInType


class Header(Parameter):
    """
    https://spec.openapis.org/oas/v3.1.0#header-object
    """

    name: Optional[str] = None  # type:ignore
    param_in: Optional[ParameterInType] = None  # type:ignore

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/5/22 8:05

import pytest
from pydantic import BaseModel

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


class MyModel(BaseModel):
    text: str


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.post('/path/')
def create_book1(body: MyModel):
    return body.text


def test_post(client):
    my_model = MyModel(text='1')
    resp = client.post("/path/", json=my_model.model_dump_json())
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/4/28 11:24
from http import HTTPStatus
from typing import Optional

from pydantic import BaseModel, Field

from flask_openapi3 import ExternalDocumentation
from flask_openapi3 import Info, Tag, Server
from flask_openapi3 import OpenAPI

info = Info(title='book API', version='1.0.0')

# Basic Authentication Sample
basic = {
    "type": "http",
    "scheme": "basic"
}
# JWT Bearer Sample
jwt = {
    "type": "http",
    "scheme": "bearer",
    "bearerFormat": "JWT"
}
# API Key Sample
api_key = {
    "type": "apiKey",
    "name": "api_key",
    "in": "header"
}
# Implicit OAuth2 Sample
oauth2 = {
    "type": "oauth2",
    "flows": {
        "implicit": {
            "authorizationUrl": "https://example.com/api/oauth/dialog",
            "scopes": {
                "write:pets": "modify pets in your account",
                "read:pets": "read your pets"
            }
        }
    }
}
security_schemes = {"jwt": jwt, "api_key": api_key, "oauth2": oauth2, "basic": basic}


class NotFoundResponse(BaseModel):
    code: int = Field(-1, description="Status Code")
    message: str = Field("Resource not found!", description="Exception Information")


app = OpenAPI(__name__, info=info, security_schemes=security_schemes, responses={404: NotFoundResponse})

book_tag = Tag(name='book', description='Some Book')
security = [
    {"jwt": []},
    {"oauth2": ["write:pets", "read:pets"]},
    {"basic": []}
]


class BookPath(BaseModel):
    bid: int = Field(..., description='book id', json_schema_extra={"deprecated": True, "example": 100})


class BookQuery(BaseModel):
    age: Optional[int] = Field(None, description='Age')
    s_list: list[str] = Field(None, alias='s_list[]', description='some array')


class BookBody(BaseModel):
    age: Optional[int] = Field(..., ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


class BookBodyWithID(BaseModel):
    bid: int = Field(..., description='book id')
    age: Optional[int] = Field(None, ge=2, le=4, description='Age')
    author: str = Field(None, min_length=2, max_length=4, description='Author')


class BookResponse(BaseModel):
    code: int = Field(0, description="Status Code")
    message: str = Field("ok", description="Exception Information")
    data: Optional[BookBodyWithID]


@app.get(
    '/book/<int:bid>',
    tags=[book_tag],
    summary='new summary',
    description='new description',
    operation_id="get_book_id",
    external_docs=ExternalDocumentation(
        url="https://www.openapis.org/",
        description="Something great got better, get excited!"),
    responses={200: BookResponse},
    security=security,
    servers=[Server(url="https://www.openapis.org/", description="openapi")]
)
def get_book(path: BookPath):
    """Get a book
    to Get some book by id, like:
    http://localhost:5000/book/3
    """
    if path.bid == 4:
        return NotFoundResponse().model_dump(), 404
    return {"code": 0, "message": "ok", "data": {"bid": path.bid, "age": 3, "author": 'no'}}


# set doc_ui False disable openapi UI
@app.get('/book', doc_ui=True, deprecated=True)
def get_books(query: BookQuery):
    """get books
    to get all books
    """
    print(query)
    return {
        "code": 0,
        "message": "ok",
        "data": [
            {"bid": 1, "age": query.age, "author": 'a1'},
            {"bid": 2, "age": query.age, "author": 'a2'}
        ]
    }


@app.post('/book', tags=[book_tag], responses={200: BookResponse})
def create_book(body: BookBody):
    print(body)
    return {"code": 0, "message": "ok"}, HTTPStatus.OK


@app.put('/book/<int:bid>', tags=[book_tag])
def update_book(path: BookPath, body: BookBody):
    print(path)
    print(body)
    return {"code": 0, "message": "ok"}


@app.delete('/book/<int:bid>', tags=[book_tag], doc_ui=False)
def delete_book(path: BookPath):
    print(path)
    return {"code": 0, "message": "ok"}


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:48
from typing import Optional, Union

from pydantic import BaseModel

from .callback import Callback
from .external_documentation import ExternalDocumentation
from .parameter import Parameter
from .reference import Reference
from .request_body import RequestBody
from .response import Response
from .security_requirement import SecurityRequirement
from .server import Server


class Operation(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#operation-object
    """

    tags: Optional[list[str]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None
    operationId: Optional[str] = None
    parameters: Optional[list[Parameter]] = None
    requestBody: Optional[Union[RequestBody, Reference]] = None
    responses: Optional[dict[str, Response]] = None
    callbacks: Optional[dict[str, Callback]] = None

    deprecated: Optional[bool] = False
    security: Optional[list[SecurityRequirement]] = None
    servers: Optional[list[Server]] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/6/21 11:23
from flask_openapi3 import Info
from flask_openapi3 import OpenAPI

info = Info(title="oauth API", version="1.0.0")

# https://spec.openapis.org/oas/v3.1.0#implicit-oauth2-sample
oauth2 = {
    "type": "oauth2",
    "flows": {
        "implicit": {
            "authorizationUrl": "https://accounts.google1.com/o/oauth2/v2/auth",
            "tokenUrl": "https://www.googleapis1.com/oauth2/v4/token",
            "scopes": {
                "write:pets": "modify pets in your account",
                "read:pets": "read your pets"
            }
        }
    }
}
security_schemes = {"oauth2": oauth2}

app = OpenAPI(__name__, info=info, security_schemes=security_schemes)

# https://github.com/swagger-api/swagger-ui/blob/0ce792c9d0965a8b6b5d75f5e1341ff6936a4cb0/docs/usage/oauth2.md
oauth_config = {
    "clientId": "xxx",
    "clientSecret": "xxx"
}
app.config["OAUTH_CONFIG"] = oauth_config

# https://spec.openapis.org/oas/v3.1.0#oauth2-security-requirement
security = [
    {"oauth2": ["write:pets", "read:pets"]}
]


@app.get("/", security=security)
def oauth():
    return "oauth"


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/4/30 14:25
import os
import re
import sys
from importlib import import_module
from typing import Optional, Union, Any, Type, Callable

from flask import Flask, Blueprint, render_template_string
from pydantic import BaseModel

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:  # pragma: no cover
    from importlib_metadata import entry_points  # type: ignore

from .blueprint import APIBlueprint
from .commands import openapi_command
from .models import APISpec
from .models import Components
from .models import ExternalDocumentation
from .models import Info
from .models import OPENAPI3_REF_PREFIX
from .models import Schema
from .models import Server
from .models import Tag
from .models import ValidationErrorModel
from .scaffold import APIScaffold
from .templates import openapi_html_string
from .types import ParametersTuple
from .types import ResponseDict
from .types import SecuritySchemesDict
from .utils import HTTPMethod
from .utils import HTTP_STATUS
from .utils import convert_responses_key_to_string
from .utils import get_model_schema
from .utils import get_operation
from .utils import get_operation_id_for_path
from .utils import get_responses
from .utils import make_validation_error_response
from .utils import parse_and_store_tags
from .utils import parse_method
from .utils import parse_parameters
from .view import APIView


class OpenAPI(APIScaffold, Flask):
    def __init__(
            self,
            import_name: str,
            *,
            info: Optional[Info] = None,
            security_schemes: Optional[SecuritySchemesDict] = None,
            responses: Optional[ResponseDict] = None,
            servers: Optional[list[Server]] = None,
            external_docs: Optional[ExternalDocumentation] = None,
            operation_id_callback: Callable = get_operation_id_for_path,
            openapi_extensions: Optional[dict[str, Any]] = None,
            validation_error_status: Union[str, int] = 422,
            validation_error_model: Type[BaseModel] = ValidationErrorModel,
            validation_error_callback: Callable = make_validation_error_response,
            doc_ui: bool = True,
            doc_prefix: str = "/openapi",
            doc_url: str = "/openapi.json",
            **kwargs: Any
    ) -> None:
        """
        OpenAPI class that provides REST API functionality along with Swagger UI and Redoc.

        Args:
            import_name: The import name for the Flask application.
            info: Information about the API (title, version, etc.).
                See https://spec.openapis.org/oas/v3.1.0#info-object.
            security_schemes: Security schemes for the API.
                See https://spec.openapis.org/oas/v3.1.0#security-scheme-object.
            responses: API responses should be either a subclass of BaseModel, a dictionary, or None.
            servers: An array of Server objects providing connectivity information to a target server.
            external_docs: External documentation for the API.
                See: https://spec.openapis.org/oas/v3.1.0#external-documentation-object.
            operation_id_callback: Callback function for custom operation ID generation.
                Receives name (str), path (str), and method (str) parameters.
                Defaults to `get_operation_id_for_path` from utils.
            openapi_extensions: Extensions to the OpenAPI Schema.
                See https://spec.openapis.org/oas/v3.1.0#specification-extensions.
            validation_error_status:
                HTTP Status of the response given when a validation error is detected by pydantic.
                Defaults to 422.
            validation_error_model: Validation error response model for OpenAPI Specification.
            validation_error_callback: Validation error response callback, the return format corresponds to
                the validation_error_model.
            doc_ui: Enable OpenAPI document UI (Swagger UI and Redoc).
                Defaults to True.
            doc_prefix: URL prefix used for OpenAPI document and UI.
                Defaults to "/openapi".
            doc_url: URL for accessing the OpenAPI specification document in JSON format.
                Defaults to "/openapi.json".
            **kwargs: Additional kwargs to be passed to Flask.
        """
        super(OpenAPI, self).__init__(import_name, **kwargs)

        # Set OpenAPI version and API information
        self.openapi_version = "3.1.0"
        self.info = info or Info(title="OpenAPI", version="1.0.0")

        # Set security schemes, responses, paths and components
        self.security_schemes = security_schemes

        # Convert key to string
        self.responses = convert_responses_key_to_string(responses or {})

        # Initialize instance variables
        self.paths: dict = dict()
        self.components_schemas: dict = dict()
        self.components = Components()

        # Initialize lists for tags and tag names
        self.tags: list[Tag] = []
        self.tag_names: list[str] = []

        # Set URL prefixes and endpoints
        self.doc_prefix = doc_prefix
        self.doc_url = doc_url

        # Set servers and external documentation
        self.severs = servers
        self.external_docs = external_docs

        # Set the operation ID callback function
        self.operation_id_callback: Callable = operation_id_callback

        # Set OpenAPI extensions
        self.openapi_extensions = openapi_extensions or {}

        # Set HTTP Response of validation errors within OpenAPI
        self.validation_error_status = str(validation_error_status)
        self.validation_error_model = validation_error_model
        self.validation_error_callback = validation_error_callback

        # Initialize the OpenAPI documentation UI
        if doc_ui:
            self._init_doc()

        # Add the OpenAPI command
        self.cli.add_command(openapi_command)  # type: ignore

        # Initialize specification JSON
        self.spec_json: dict = {}
        self.spec = APISpec(
            openapi=self.openapi_version,
            info=self.info,
            paths=self.paths
        )

    def _init_doc(self) -> None:
        """
        Provide Swagger UI, Redoc, and Rapidoc
        """
        _here = os.path.dirname(__file__)
        template_folder = os.path.join(_here, "templates")
        static_folder = os.path.join(template_folder, "static")

        # Create the blueprint for OpenAPI documentation
        blueprint = Blueprint(
            "openapi",
            __name__,
            url_prefix=self.doc_prefix,
            template_folder=template_folder,
            static_folder=static_folder
        )

        # Add the API documentation URL rule
        blueprint.add_url_rule(
            rule=self.doc_url,
            endpoint="doc_url",
            view_func=lambda: self.api_doc
        )

        ui_templates = []
        # Iterate over all entry points in the "flask_openapi3.plugins" group
        for entry_point in entry_points(group="flask_openapi3.plugins"):
            try:
                module_path = entry_point.value
                module_name, class_name = module_path.rsplit(".", 1)
                module = import_module(module_name)
                plugin_class = getattr(module, class_name)
                plugin_register = plugin_class.register
                plugin_name = plugin_class.name
                plugin_display_name = plugin_class.display_name
                bp = plugin_register(doc_url=self.doc_url.lstrip("/"))
                self.register_blueprint(bp, url_prefix=self.doc_prefix)
                ui_templates.append({"name": plugin_name, "display_name": plugin_display_name})
            except (ModuleNotFoundError, AttributeError):  # pragma: no cover
                import traceback
                print(f"Warning: plugin '{entry_point.value}' registration failed.")
                traceback.print_exc()

        # Add URL rule for the home page
        blueprint.add_url_rule(
            rule="/",
            endpoint="openapi",
            view_func=lambda: render_template_string(
                self.config.get("OPENAPI_HTML_STRING") or openapi_html_string,
                ui_templates=ui_templates
            )
        )

        # Register the blueprint with the Flask application
        self.register_blueprint(blueprint)

    @property
    def api_doc(self) -> dict:
        """
        Generate the OpenAPI specification JSON.

        Returns:
            The OpenAPI specification JSON as a dictionary.

        """
        if self.spec_json:
            return self.spec_json

        self.generate_spec_json()

        return self.spec_json

    def generate_spec_json(self):
        self.spec.openapi = self.openapi_version
        self.spec.info = self.info
        self.spec.paths = self.paths

        if self.severs:
            self.spec.servers = self.severs

        if self.external_docs:
            self.spec.externalDocs = self.external_docs

        # Set tags
        if self.tags:
            self.spec.tags = self.tags

        # Add ValidationErrorModel to components schemas
        schema = get_model_schema(self.validation_error_model)
        self.components_schemas[self.validation_error_model.__name__] = Schema(**schema)

        # Parse definitions
        definitions = schema.get("$defs", {})
        for name, value in definitions.items():
            self.components_schemas[name] = Schema(**value)

        # Set components
        self.components.schemas = self.components_schemas
        self.components.securitySchemes = self.security_schemes
        self.spec.components = self.components

        # Convert spec to JSON
        self.spec_json = self.spec.model_dump(mode="json", by_alias=True, exclude_unset=True, warnings=False)

        # Update with OpenAPI extensions
        self.spec_json.update(**self.openapi_extensions)

        # Handle validation error response
        for rule, path_item in self.spec_json["paths"].items():
            for http_method, operation in path_item.items():
                if operation.get("parameters") is None and operation.get("requestBody") is None:
                    continue
                if not operation.get("responses"):
                    operation["responses"] = {}
                if operation["responses"].get(self.validation_error_status):
                    continue
                operation["responses"][self.validation_error_status] = {
                    "description": HTTP_STATUS[self.validation_error_status],
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "array",
                                "items": {"$ref": f"{OPENAPI3_REF_PREFIX}/{self.validation_error_model.__name__}"}
                            }
                        }
                    }
                }

    def register_api(self, api: APIBlueprint, **options: Any) -> None:
        """
        Register an APIBlueprint.

        Args:
            api: The APIBlueprint instance to register.

            options: Additional keyword arguments are passed to :class:`~flask.blueprints.BlueprintSetupState`.
            They can be accessed in :meth:`~flask.Blueprint.record` callbacks.
            url_prefix, Blueprint routes will be prefixed with this.
            subdomain, Blueprint routes will match on this subdomain.
            url_defaults, Blueprint routes will use these default values for view arguments.

        """
        for tag in api.tags:
            if tag.name not in self.tag_names:
                # Append tag to the list of tags
                self.tags.append(tag)

                # Append tag name to the list of tag names
                self.tag_names.append(tag.name)

        # Update paths with the APIBlueprint's paths
        url_prefix = options.get("url_prefix")
        if url_prefix and api.url_prefix and url_prefix != api.url_prefix:
            api.paths = {url_prefix + k.removeprefix(api.url_prefix): v for k, v in api.paths.items()}
        elif url_prefix and not api.url_prefix:
            api.paths = {url_prefix.rstrip("/") + "/" + k.lstrip("/"): v for k, v in api.paths.items()}
        self.paths.update(**api.paths)

        # Update component schemas with the APIBlueprint's component schemas
        self.components_schemas.update(**api.components_schemas)

        # Register the APIBlueprint with the current instance
        self.register_blueprint(api, **options)

    def register_api_view(
            self,
            api_view: APIView,
            url_prefix: Optional[str] = None,
            view_kwargs: Optional[dict[Any, Any]] = None
    ) -> None:
        """
        Register APIView

        Args:
            api_view: The APIView instance to register.
            url_prefix: A path to prepend to all the APIView's urls
            view_kwargs: Additional keyword arguments to pass to the API views.
        """
        if view_kwargs is None:
            view_kwargs = {}

        # Iterate through tags of the APIView
        for tag in api_view.tags:
            if tag.name not in self.tag_names:
                # Append tag to the list of tags
                self.tags.append(tag)

                # Append tag name to the list of tag names
                self.tag_names.append(tag.name)

        # Update paths with the APIView's paths
        if url_prefix and api_view.url_prefix and url_prefix != api_view.url_prefix:
            api_view.paths = {url_prefix + k.removeprefix(api_view.url_prefix): v for k, v in api_view.paths.items()}
        elif url_prefix and not api_view.url_prefix:
            api_view.paths = {url_prefix.rstrip("/") + "/" + k.lstrip("/"): v for k, v in api_view.paths.items()}
        self.paths.update(**api_view.paths)

        # Update component schemas with the APIView's component schemas
        self.components_schemas.update(**api_view.components_schemas)

        # Register the APIView with the current instance
        api_view.register(self, url_prefix=url_prefix, view_kwargs=view_kwargs)

    def _add_url_rule(
            self,
            rule,
            endpoint=None,
            view_func=None,
            provide_automatic_options=None,
            **options,
    ) -> None:
        self.add_url_rule(rule, endpoint, view_func, provide_automatic_options, **options)

    def _collect_openapi_info(
            self,
            rule: str,
            func: Callable,
            *,
            tags: Optional[list[Tag]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            external_docs: Optional[ExternalDocumentation] = None,
            operation_id: Optional[str] = None,
            responses: Optional[ResponseDict] = None,
            deprecated: Optional[bool] = None,
            security: Optional[list[dict[str, list[Any]]]] = None,
            servers: Optional[list[Server]] = None,
            openapi_extensions: Optional[dict[str, Any]] = None,
            doc_ui: bool = True,
            method: str = HTTPMethod.GET
    ) -> ParametersTuple:
        """
        Collects OpenAPI specification information for Flask routes and view functions.

        Args:
            rule: Flask route.
            func: Flask view_func.
            tags: Adds metadata to a single tag.
            summary: A short summary of what the operation does.
            description: A verbose explanation of the operation behavior.
            external_docs: Additional external documentation for this operation.
            operation_id: Unique string used to identify the operation.
            responses: API responses should be either a subclass of BaseModel, a dictionary, or None.
            deprecated: Declares this operation to be deprecated.
            security: A declaration of which security mechanisms can be used for this operation.
            servers: An alternative server array to service this operation.
            openapi_extensions: Allows extensions to the OpenAPI Schema.
            doc_ui: Declares this operation to be shown. Default to True.
            method: HTTP method for the operation. Defaults to GET.
        """
        if doc_ui is True:
            # Convert key to string
            new_responses = convert_responses_key_to_string(responses or {})

            # Global response: combine API responses
            combine_responses = {**self.responses, **new_responses}

            # Create operation
            operation = get_operation(
                func,
                summary=summary,
                description=description,
                openapi_extensions=openapi_extensions
            )
            # Set external docs
            if external_docs:
                operation.externalDocs = external_docs

            # Unique string used to identify the operation.
            operation.operationId = operation_id or self.operation_id_callback(
                name=func.__name__, path=rule, method=method
            )

            # Only set `deprecated` if True, otherwise leave it as None
            if deprecated is not None:
                operation.deprecated = deprecated

            # Add security
            if security:
                operation.security = security

            # Add servers
            if servers:
                operation.servers = servers

            # Store tags
            parse_and_store_tags(tags or [], self.tags, self.tag_names, operation)

            # Parse response
            get_responses(combine_responses, self.components_schemas, operation)

            # Convert a route parameter format from /pet/<petId> to /pet/{petId}
            uri = re.sub(r"<([^<:]+:)?", "{", rule).replace(">", "}")

            # Parse method
            parse_method(uri, method, self.paths, operation)

            # Parse parameters
            return parse_parameters(func, components_schemas=self.components_schemas, operation=operation)
        else:
            return parse_parameters(func, doc_ui=False)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/9/29 10:36
import pytest
from pydantic import BaseModel

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


class BookQuery(BaseModel):
    age: list[int] = [1, 2]


class BookForm(BaseModel):
    age: list[float] = [3, 4]


@app.get("/query")
def api_query(query: BookQuery):
    assert query.age == [1, 2]
    return {"code": 0, "message": "ok"}


@app.post("/form")
def api_form(form: BookForm):
    assert form.age == [3, 4]
    return {"code": 0, "message": "ok"}


def test_query(client):
    client.get("/query")


def test_form(client):
    client.post("/form")



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/12/19 10:34

import pytest

from flask_openapi3 import Info, Tag
from flask_openapi3 import OpenAPI, APIBlueprint

info = Info(title="book API", version="1.0.0")

app = OpenAPI(__name__, info=info)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()
    return client


api1 = APIBlueprint("book1", __name__)


@api1.get('/book', tags=[Tag(name="book")])
def get_book():
    ...  # pragma: no cover


api2 = APIBlueprint("book2", __name__)


@api2.get('/book2', tags=[Tag(name="book")])
def get_book2():
    ...  # pragma: no cover


app.register_api(api1)
app.register_api(api2)


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert resp.status_code == 200
    tags = _json["tags"]
    news_tags = []
    for tag in tags:
        if tag not in news_tags:
            news_tags.append(tag)
    assert news_tags == tags



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:57
from typing import Optional

from pydantic import BaseModel, Field


class ServerVariable(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#server-variable-object
    """

    enum: Optional[list[str]] = Field(None, min_length=1)
    default: str
    description: Optional[str] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:47
from typing import Optional

from pydantic import BaseModel

from .oauth_flow import OAuthFlow


class OAuthFlows(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#oauth-flows-object
    """

    implicit: Optional[OAuthFlow] = None
    password: Optional[OAuthFlow] = None
    clientCredentials: Optional[OAuthFlow] = None
    authorizationCode: Optional[OAuthFlow] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/6/30 10:12
import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, FileStorage

app = OpenAPI(__name__)
app.config["TESTING"] = True


class UploadFilesForm(BaseModel):
    file: FileStorage
    str_list: list[str]

    model_config = dict(
        openapi_extra={
            # "example": {"a": 123},
            "examples": {
                "Example 01": {
                    "summary": "An example",
                    "value": {
                        "file": "Example-01.jpg",
                        "str_list": ["a", "b", "c"]
                    }
                },
                "Example 02": {
                    "summary": "Another example",
                    "value": {
                        "str_list": ["1", "2", "3"]
                    }
                }
            }
        }
    )


class BookBody(BaseModel):
    age: int
    author: str

    model_config = dict(
        openapi_extra={
            "description": "This is post RequestBody",
            "example": {"age": 12, "author": "author1"},
            "examples": {
                "example1": {
                    "summary": "example summary1",
                    "description": "example description1",
                    "value": {
                        "age": 24,
                        "author": "author2"
                    }
                },
                "example2": {
                    "summary": "example summary2",
                    "description": "example description2",
                    "value": {
                        "age": 48,
                        "author": "author3"
                    }
                }

            }}
    )


class MessageResponse(BaseModel):
    message: str = Field(..., description="The message")
    metadata: dict[str, str] = Field(alias="metadata_")

    model_config = dict(
        by_alias=False,
        openapi_extra={
            # "example": {"message": "aaa"},
            "examples": {
                "example1": {
                    "summary": "example1 summary",
                    "value": {
                        "message": "bbb"
                    }
                },
                "example2": {
                    "summary": "example2 summary",
                    "value": {
                        "message": "ccc"
                    }
                }
            }
        }
    )


@app.post("/form")
def api_form(form: UploadFilesForm):
    print(form)  # pragma: no cover


@app.post("/body", responses={"200": MessageResponse})
def api_error_json(body: BookBody):
    print(body)  # pragma: no cover


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert resp.status_code == 200
    assert _json["paths"]["/form"]["post"]["requestBody"]["content"]["multipart/form-data"]["examples"] == \
           {
               "Example 01": {
                   "summary": "An example",
                   "value": {
                       "file": "Example-01.jpg",
                       "str_list": ["a", "b", "c"]
                   }
               },
               "Example 02": {
                   "summary": "Another example",
                   "value": {
                       "str_list": ["1", "2", "3"]
                   }
               }
           }
    assert _json["paths"]["/body"]["post"]["requestBody"]["description"] == "This is post RequestBody"
    assert _json["paths"]["/body"]["post"]["requestBody"]["content"]["application/json"]["example"] == \
           {"age": 12, "author": "author1"}
    assert _json["paths"]["/body"]["post"]["responses"]["200"]["content"]["application/json"]["examples"] == \
           {
               "example1": {
                   "summary": "example1 summary",
                   "value": {
                       "message": "bbb"
                   }
               },
               "example2": {
                   "summary": "example2 summary",
                   "value": {
                       "message": "ccc"
                   }
               }
           }
    assert _json["components"]["schemas"]["MessageResponse"]["properties"].get("metadata") is not None



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/9 15:25
from http import HTTPStatus
from typing import Union, Type, Any, Optional

from pydantic import BaseModel

from .models import RawModel
from .models import SecurityScheme

_ResponseDictValue = Union[Type[BaseModel], dict[Any, Any], None]

ResponseDict = dict[Union[str, int, HTTPStatus], _ResponseDictValue]

ResponseStrKeyDict = dict[str, _ResponseDictValue]

SecuritySchemesDict = dict[str, Union[SecurityScheme, dict[str, Any]]]

ParametersTuple = tuple[
    Optional[Type[BaseModel]],
    Optional[Type[BaseModel]],
    Optional[Type[BaseModel]],
    Optional[Type[BaseModel]],
    Optional[Type[BaseModel]],
    Optional[Type[BaseModel]],
    Optional[Type[RawModel]]
]



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/12/26 15:07

from pydantic import BaseModel, Field

from flask_openapi3 import Info, Tag
from flask_openapi3 import OpenAPI

info = Info(title='header API', version='1.0.0')
app = OpenAPI(__name__, info=info)

book_tag = Tag(name='book', description='Some Book')


class Headers(BaseModel):
    hello: str = Field("what's up", max_length=12, description='sds')
    # required
    # hello: str = Field(..., max_length=12, description='sds')
    x_hello: str = Field(..., max_length=12, description='Header with alias to support dash', alias="x-hello")


@app.get('/book', tags=[book_tag])
def get_book(header: Headers):
    print(header)
    return header.hello


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/4/30 10:14

from .__version__ import __version__
from .blueprint import APIBlueprint
from .models import APISpec
from .models import Components
from .models import Contact
from .models import Discriminator
from .models import Encoding
from .models import Example
from .models import ExternalDocumentation
from .models import FileStorage
from .models import Header
from .models import Info
from .models import License
from .models import Link
from .models import MediaType
from .models import OAuthConfig
from .models import OAuthFlow
from .models import OAuthFlows
from .models import Operation
from .models import Parameter
from .models import ParameterInType
from .models import PathItem
from .models import RawModel
from .models import Reference
from .models import RequestBody
from .models import Response
from .models import Schema
from .models import SecurityScheme
from .models import Server
from .models import ServerVariable
from .models import StyleValues
from .models import Tag
from .models import UnprocessableEntity
from .models import ValidationErrorModel
from .models import XML
from .openapi import OpenAPI
from .view import APIView



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/4/28 10:58
from typing import Optional

from pydantic import BaseModel

from .contact import Contact
from .license import License


class Info(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#info-object
    """

    title: str
    summary: Optional[str] = None
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None
    version: str

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/11/4 14:41

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import APIView
from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True

api_view = APIView(url_prefix="/api/v1")


class BookPath(BaseModel):
    id: int = Field(..., description="book ID")


@api_view.route("/book")
class BookListAPIView:
    def __init__(self, view_kwargs=None):
        self.a = view_kwargs.get("a")

    @api_view.doc(summary="get book list")
    def get(self):
        return {"a": self.a}


@api_view.route("/book/<id>")
class BookAPIView:
    def __init__(self, view_kwargs=None):
        self.b = view_kwargs.get("b")

    @api_view.doc(summary="get book list")
    async def get(self, path: BookPath):
        print(path)
        return {"b": self.b}


app.register_api_view(
    api_view,
    view_kwargs={
        "a": 1,
        "b": 2
    }
)


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_get_list_view_kwargs(client):
    resp = client.get("/api/v1/book")
    assert resp.status_code == 200

    assert resp.json["a"] == 1


def test_get_view_kwargs(client):
    resp = client.get("/api/v1/book/1")
    assert resp.status_code == 200

    assert resp.json["b"] == 2



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:41
from typing import TYPE_CHECKING, Optional, Union

from pydantic import BaseModel

from .reference import Reference

if TYPE_CHECKING:  # pragma: no cover
    from .header import Header
else:
    Header = "Header"


class Encoding(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#encoding-object
    """

    contentType: Optional[str] = None
    headers: Optional[dict[str, Union[Header, Reference]]] = None
    style: Optional[str] = None
    explode: Optional[bool] = None
    allowReserved: bool = False

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/11 13:37

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)


@app.route('/')
def hello_world():
    return 'Hello, World!'


if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:46
from typing import Any, Optional, Union

from pydantic import BaseModel, Field

from .encoding import Encoding
from .example import Example
from .reference import Reference
from .schema import Schema


class MediaType(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#media-type-object
    """

    media_type_schema: Optional[Union[Reference, Schema]] = Field(default=None, alias="schema")
    example: Optional[Any] = None
    examples: Optional[dict[str, Union[Example, Reference]]] = None
    encoding: Optional[dict[str, Encoding]] = None

    model_config = {
        "extra": "allow"
    }



from typing import Optional

from pydantic import BaseModel

from .external_documentation import ExternalDocumentation


class Tag(BaseModel):
    """
    https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#tag-object
    """

    name: str
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2021/5/11 14:03
from pydantic import BaseModel, Field

from flask_openapi3 import OpenAPI, FileStorage

app = OpenAPI(__name__)


class UploadFileForm(BaseModel):
    file: FileStorage
    file_type: str = Field(None, description="File Type")


class UploadFilesForm(BaseModel):
    files: list[FileStorage]
    str_list: list[str]
    int_list: list[int]


@app.post('/upload/file')
def upload_file(form: UploadFileForm):
    print(form.file.filename)
    print(form.file_type)
    form.file.save('test.jpg')
    return {"code": 0, "message": "ok"}


@app.post('/upload/files')
def upload_files(form: UploadFilesForm):
    print(form.files)
    print(form.str_list)
    print(form.int_list)
    return {"code": 0, "message": "ok"}


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/2/27 16:53
from typing import Any

import pytest
from pydantic import BaseModel, RootModel

from flask_openapi3 import OpenAPI, Tag

app = OpenAPI(__name__)
app.config["TESTING"] = True


class Sellout(BaseModel):
    a: str
    b: int


class SelloutList(RootModel):
    root: list[Sellout]


class SelloutDict(RootModel):
    root: dict[str, Sellout]


class SelloutDict2(RootModel):
    root: dict[Any, Any]


class SelloutDict3(BaseModel):
    model_config = {
        "extra": "allow",
    }


@pytest.fixture
def client():
    client = app.test_client()

    return client


@app.post('/api/v1/sellouts',
          tags=[Tag(name='Sellout', description='Loren.')],
          responses={'200': SelloutList}
          )
def post_sellout(body: SelloutList):
    print(body)
    return body.model_dump_json()


@app.post('/api/v2/sellouts',
          tags=[Tag(name='Sellout', description='Loren.')],
          responses={'200': SelloutDict}
          )
def post_sellout2(body: SelloutDict):
    print(body)
    return body.model_dump_json()


@app.post('/api/v3/sellouts',
          tags=[Tag(name='Sellout', description='Loren.')]
          )
def post_sellout3(body: SelloutDict2):
    print(body)
    return body.model_dump_json()


@app.post('/api/v4/sellouts',
          tags=[Tag(name='Sellout', description='Loren.')]
          )
def post_sellout4(body: SelloutDict3):
    print(body)
    return body.model_dump_json()


def test_v1_sellouts(client):
    resp = client.post("/api/v1/sellouts", json=[{"a": "string", "b": 0}])
    assert resp.status_code == 200


def test_v2_sellouts(client):
    resp = client.post("/api/v2/sellouts", json={"additionalProp1": {"a": "string", "b": 0}})
    assert resp.status_code == 200


def test_v3_sellouts(client):
    resp = client.post("/api/v3/sellouts", json={"a": 11, "b": 23, "c": {"cc": 33, "ccc": 333}})
    assert resp.status_code == 200


def test_v4_sellouts(client):
    resp = client.post("/api/v4/sellouts", json={"a": 11, "b": 23, "c": {"cc": 33, "ccc": 333}})
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:54
from typing import Optional, Union

from pydantic import BaseModel

from .header import Header
from .link import Link
from .media_type import MediaType
from .reference import Reference


class Response(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#response-object
    """

    description: str
    headers: Optional[dict[str, Union[Header, Reference]]] = None
    content: Optional[dict[str, MediaType]] = None
    links: Optional[dict[str, Union[Link, Reference]]] = None

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:58
from typing import Optional

from pydantic import BaseModel


class XML(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#xml-object
    """

    name: Optional[str] = None
    namespace: Optional[str] = None
    prefix: Optional[str] = None
    attribute: bool = False
    wrapped: bool = False

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/4 9:53
from typing import Optional

from pydantic import BaseModel

from .media_type import MediaType


class RequestBody(BaseModel):
    """
    https://spec.openapis.org/oas/v3.1.0#request-body-object
    """

    description: Optional[str] = None
    content: dict[str, MediaType]
    required: Optional[bool] = True

    model_config = {
        "extra": "allow"
    }



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2025/1/15 9:58
import pytest

from flask_openapi3 import APIBlueprint, OpenAPI, APIView

app = OpenAPI(__name__, )
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


api1 = APIBlueprint("/book1", __name__, url_prefix="/api")
api2 = APIBlueprint("/book2", __name__)


@api1.get("/book")
def create_book1():
    return "ok"


@api2.get("/book")
def create_book2():
    return "ok"


app.register_api(api1, url_prefix="/api1")
app.register_api(api2, url_prefix="/api2")

api_view1 = APIView(url_prefix="/api")
api_view2 = APIView()


@api_view1.route("/book")
class BookAPIView:
    @api_view1.doc(summary="get book")
    def get(self):
        return "ok"


@api_view2.route("/book")
class BookAPIView2:
    @api_view2.doc(summary="get book")
    def get(self, ):
        return "ok"


app.register_api_view(api_view1, url_prefix="/api3")
app.register_api_view(api_view2, url_prefix="/api4")


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert "/api1/book" in _json["paths"].keys()
    assert "/api2/book" in _json["paths"].keys()
    assert "/api3/book" in _json["paths"].keys()
    assert "/api4/book" in _json["paths"].keys()



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/7/3 9:44
import os

import pytest

from flask_openapi3 import OpenAPI
from flask_openapi3.commands import openapi_command


@pytest.fixture
def app():
    app = OpenAPI(__name__)
    return app


@pytest.fixture
def cli_runner(app):
    return app.test_cli_runner()


def test_openapi_command(app, cli_runner):
    result = cli_runner.invoke(openapi_command, ("--output", "test.json"))
    os.remove("test.json")

    assert result.exit_code == 0
    assert "Saved to test.json." in result.output

    result = cli_runner.invoke(openapi_command, ("--format", "yaml"))

    assert result.exit_code == 0



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2022/4/2 9:09

import pytest
from pydantic import BaseModel

from flask_openapi3 import APIBlueprint, OpenAPI, Tag

app = OpenAPI(__name__)

api = APIBlueprint('book', __name__, url_prefix='/api/book/<name>')
api_english = APIBlueprint('english', __name__)
api_chinese = APIBlueprint('chinese', __name__)


class BookPath(BaseModel):
    name: str


@api_english.post('/english')
def create_english_book(path: BookPath):
    return {"message": "english", "name": path.name}


@api_chinese.post('/chinese', tags=[Tag(name="chinese")])
def create_chinese_book(path: BookPath):
    return {"message": "chinese", "name": path.name}


# register nested api
api.register_api(api_english)
api.register_api(api_chinese)
# register api
app.register_api(api)


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json == app.api_doc


def test_post_english(client):
    resp = client.post("/api/book/name1/english")
    assert resp.status_code == 200
    assert resp.json == {"message": "english", "name": "name1"}


def test_post_chinese(client):
    resp = client.post("/api/book/name2/chinese")
    assert resp.status_code == 200
    assert resp.json == {"message": "chinese", "name": "name2"}



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/7/21 10:32
import pytest
from flask import make_response, current_app
from pydantic import BaseModel, Field, ValidationError

from flask_openapi3 import OpenAPI


class GenericTracebackError(BaseModel):
    location: str = Field(..., json_schema_extra={"example": "GenericError.py"})
    line: int = Field(..., json_schema_extra={"example": 1})
    method: str = Field(..., json_schema_extra={"example": "GenericError"})
    message: str = Field(..., json_schema_extra={"example": "400:Bad Request"})


class ValidationErrorModel(BaseModel):
    code: str
    message: str
    more_info: list[GenericTracebackError] = Field(..., json_schema_extra={"example": [GenericTracebackError(
        location="GenericError.py",
        line=1,
        method="GenericError",
        message="400:Bad Request")]})


def validation_error_callback(e: ValidationError):
    validation_error_object = ValidationErrorModel(
        code="400",
        message=e.json(),
        more_info=[GenericTracebackError(
            location="GenericError.py",
            line=1,
            method="GenericError",
            message="400:Bad Request"
        )]
    )
    response = make_response(validation_error_object.model_dump_json())
    response.headers["Content-Type"] = "application/json"
    response.status_code = getattr(current_app, "validation_error_status", 422)
    return response


app = OpenAPI(
    __name__,
    validation_error_status=400,
    validation_error_model=ValidationErrorModel,
    validation_error_callback=validation_error_callback
)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()
    return client


class BookQuery(BaseModel):
    age: int = Field(None, description="Age")


@app.get("/query")
def api_query(query: BookQuery):
    print(query)  # pragma: no cover


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    assert resp.status_code == 200
    assert resp.json["components"]["schemas"].get("GenericTracebackError") is not None


def test_api_query(client):
    resp = client.get("/query?age=abc")
    print(resp.json)
    assert resp.status_code == 400



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/2/5 14:54

from enum import Enum

import pytest
from pydantic import BaseModel, Field

from flask_openapi3 import Info
from flask_openapi3 import OpenAPI

app = OpenAPI(
    __name__,
    info=Info(title="Enum demo", version="1.0.0")
)

app.config["TESTING"] = True


class Language(str, Enum):
    cn = 'Chinese'
    en = 'English'


class LanguagePath(BaseModel):
    language: Language = Field(..., description='Language')


@app.get('/<language>')
def get_enum(path: LanguagePath):
    print(path)
    return {}


@pytest.fixture
def client():
    client = app.test_client()

    return client


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    _json = resp.json
    assert resp.status_code == 200
    assert _json["components"]["schemas"].get("Language") is not None

    resp = client.get("/English")
    assert resp.status_code == 200



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/4/23 14:31


class BasePlugin:
    name = None
    display_name = None

    @classmethod
    def register(cls, doc_url: str) -> None:
        """
        Register the plugin.

        Args:
            doc_url (str): The API doc url.
        """



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/4/7 11:13
from enum import Enum


class SecuritySchemeInType(str, Enum):
    """The place Parameters can be put when calling an Endpoint"""

    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2024/1/28 16:38
from functools import cached_property

import pytest
from pydantic import BaseModel, Field, computed_field

from flask_openapi3 import OpenAPI

app = OpenAPI(__name__)
app.config["TESTING"] = True


@pytest.fixture
def client():
    client = app.test_client()

    return client


class User(BaseModel):
    firstName: str = Field(title="First Name")
    lastName: str

    @computed_field(title="Display Name")
    @cached_property
    def display_name(self) -> str:
        return f"{self.firstName} {self.lastName}"  # pragma: no cover


@app.get("/user", responses={200: User})
def get_book():
    return "ok"  # pragma: no cover


def test_openapi(client):
    resp = client.get("/openapi/openapi.json")
    import pprint
    pprint.pprint(resp.json)
    assert resp.json["components"]["schemas"]["User"]["properties"].get("display_name") is not None



# -*- coding: utf-8 -*-
# @Author  : llc
# @Time    : 2023/5/31 14:24
from flask_openapi3 import OpenAPI

app = OpenAPI(__name__, openapi_extensions={
    "x-google-endpoints": [
        {
            "name": "my-cool-api.endpoints.my-project-id.cloud.goog",
            "allowCors": True
        }
    ]
})

openapi_extensions = {
    "x-google-backend": {
        "address": "https://<NODE_SERVICE_ID>-<HASH>.a.run.app",
        "protocol": "h2"
    },
    "x-aperture-labs-portal": "blue"
}


@app.get("/", openapi_extensions=openapi_extensions)
def hello():
    return "ok"


if __name__ == "__main__":
    app.run(debug=True)
